"""
Manager class to ease dealing with the chardet encoding detection library 'chardet'.
Each instance of this classes managed a chardet.detect_all() scan on a single set of bytes.
"""
from collections import namedtuple
from numbers import Number
from operator import attrgetter
from typing import List

import chardet
from rich import box
from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from pdfalyzer.detection.encoding_assessment import ENCODING, EncodingAssessment
from pdfalyzer.helpers.rich_text_helper import console
from pdfalyzer.util.logging import log


# TODO: move to config.py?
MIN_BYTES_FOR_ENCODING_DETECTION = 9
CONFIDENCE_SCORE_RANGE = range(0, 100)


class EncodingDetector:
    # 10 as in 10%, 0.02, etc.  Encodings w/confidences below this will not be displayed in the decoded table
    force_display_threshold = 20.0

    # At what chardet.detect() confidence % should we force a decode with an obscure encoding?
    force_decode_threshold = 50.0

    def __init__(self, _bytes: bytes) -> None:
        self.bytes = _bytes
        self.bytes_len = len(_bytes)
        self.table = _empty_chardet_results_table()

        if not self.has_enough_bytes():
            log.debug(f"{self.bytes_len} is not enough bytes to run chardet.detect()")
            self._set_empty_results()
            self.has_any_idea = None  # not false!
            return

        # Unique by encoding, ignoring language.  Ordered from highest to lowest confidence
        self.unique_assessments = []
        self.raw_chardet_assessments = chardet.detect_all(self.bytes, ignore_threshold=True)

        if len(self.raw_chardet_assessments) == 1 and self.raw_chardet_assessments[0][ENCODING] is None:
            log.info(f"chardet.detect() has no idea what the encoding is, result: {self.raw_chardet_assessments}")
            self._set_empty_results()
            self.has_any_idea = False
            return

        self.has_any_idea = True
        self.assessments = [EncodingAssessment(a) for a in self.raw_chardet_assessments]
        self._uniquify_results_and_build_table()
        self.force_decode_assessments = self.assessments_above_confidence(type(self).force_decode_threshold)
        self.force_display_assessments = self.assessments_above_confidence(type(self).force_display_threshold)

    def get_encoding_assessment(self, encoding) -> EncodingAssessment:
        """If chardet produced one, return it, otherwise return a dummy node with confidence of 0"""
        assessment = next((r for r in self.unique_assessments if r.encoding == encoding), None)
        return assessment or EncodingAssessment.dummy_encoding_assessment(encoding)

    def has_enough_bytes(self) -> bool:
        return self.bytes_len >= MIN_BYTES_FOR_ENCODING_DETECTION

    def assessments_above_confidence(self, cutoff: float) -> List[EncodingAssessment]:
        return [a for a in self.unique_assessments if a.confidence >= cutoff]

    def __rich__(self) -> Padding:
        return Padding(self.table, (0, 0, 0, 20))

    def _uniquify_results_and_build_table(self) -> None:
        """Keep the highest result per encoding, ignoring the language chardet has indicated"""
        already_seen_encodings = {}

        for i, result in enumerate(self.assessments):
            self.table.add_row(f"{i + 1}", result.encoding_text, result.confidence_text)

            # self.unique_assessments retains one result per encoding possibility (the highest confidence one)
            # Some encodings are not language specific and for those we don't care about the language
            if result.encoding not in already_seen_encodings:
                self.unique_assessments.append(result)
                already_seen_encodings[result.encoding] = result
            else:
                log.debug(f"Skipping chardet result {result}: we already saw {already_seen_encodings[result.encoding]})")

        self.unique_assessments.sort(key=attrgetter('confidence'), reverse=True)

    def _set_empty_results(self) -> None:
        self.assessments = []
        self.unique_assessments = []
        self.raw_chardet_assessments = []
        self.force_decode_assessments = []
        self.force_display_assessments = []


def _empty_chardet_results_table():
    """Returns a fresh table"""
    table = Table(
        'Rank', 'Encoding', 'Confidence',
        title='chardet encoding detection',
        title_style='color(153) italic dim',
        header_style='off_white',
        style='dim',
        box=box.SIMPLE,
        show_edge=False,
        collapse_padding=True)

    table.columns[0].justify = 'right'
    table.columns
    return table



import re
from io import StringIO

from rich.console import Console
from rich.text import Text

from pdfalyzer.binary.bytes_match import BytesMatch
from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.detection.constants.character_encodings import NEWLINE_BYTE
from pdfalyzer.helpers.rich_text_helper import BYTES_HIGHLIGHT, console
from pdfalyzer.util.logging import log

HEX_CHARS_PER_GROUP = 8
HEX_GROUPS_PER_LINE = 4


def get_bytes_before_and_after_match(_bytes: bytes, match: re.Match, num_before=None, num_after=None) -> bytes:
    """
    Get all bytes from num_before the start of the sequence up until num_after the end of the sequence
    num_before and num_after will both default to the env var/CLI options having to do with surrounding
    bytes. If only num_before is provided then num_after will use it as a default.
    """
    return get_bytes_surrounding_range(_bytes, match.start(), match.end(), num_before, num_after)


def get_bytes_surrounding_range(_bytes: bytes, start_idx: int, end_idx: int, num_before=None, num_after=None) -> bytes:
    num_after = num_after or num_before or PdfalyzerConfig.NUM_SURROUNDING_BYTES
    num_before = num_before or PdfalyzerConfig.NUM_SURROUNDING_BYTES
    start_idx = max(start_idx - num_before, 0)
    end_idx = min(end_idx + num_after, len(_bytes))
    return _bytes[start_idx:end_idx]


def clean_byte_string(bytes_array: bytes) -> str:
    """Gives you a string representation of bytes w/no cruft e.g. '\x80\nx44' instead of "b'\x80\nx44'"."""
    byte_printer = Console(file=StringIO())
    byte_printer.out(bytes_array, end='')
    bytestr = byte_printer.file.getvalue()

    if bytestr.startswith("b'"):
        bytestr = bytestr.removeprefix("b'").removesuffix("'")
    elif bytestr.startswith('b"'):
        bytestr = bytestr.removeprefix('b"').removesuffix('"')
    else:
        raise RuntimeError(f"Unexpected byte string {bytestr}")

    return bytestr


def rich_text_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    """Print raw bytes to a Text object, highlighing the bytes in the bytes_match BytesMatch"""
    surrounding_bytes_str = clean_byte_string(_bytes)
    highlighted_bytes_str = clean_byte_string(bytes_match.bytes)
    highlighted_bytes_str_length = len(highlighted_bytes_str)
    highlight_idx = _find_str_rep_of_bytes(surrounding_bytes_str, highlighted_bytes_str, bytes_match)

    txt = Text(surrounding_bytes_str[:highlight_idx], style='grey')
    matched_bytes_str = surrounding_bytes_str[highlight_idx:highlight_idx + highlighted_bytes_str_length]
    txt.append(matched_bytes_str, style=bytes_match.highlight_style)
    txt.append(surrounding_bytes_str[highlight_idx + highlighted_bytes_str_length:], style='grey')
    return txt


def hex_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    hex_str = Text(' '.join([hex(b).removeprefix('0x').rjust(2, '0') for i, b in enumerate(_bytes)]), style='grey')
    highlight_start_idx = bytes_match.highlight_start_idx * 3
    highlight_end_idx = bytes_match.highlight_end_idx * 3
    hex_str.stylize(bytes_match.highlight_style, highlight_start_idx, highlight_end_idx)
    lines = hex_str.wrap(console, HEX_CHARS_PER_GROUP * HEX_GROUPS_PER_LINE * 3)
    return Text("\n").join([Text('  ').join(line.wrap(console, HEX_CHARS_PER_GROUP * 3)) for line in lines])


def print_bytes(bytes_array: bytes, style=None) -> None:
    """Convert bytes to a string representation and print to console"""
    for line in bytes_array.split(NEWLINE_BYTE):
        console.print(clean_byte_string(line), style=style or 'bytes')


def _find_str_rep_of_bytes(surrounding_bytes_str: str, highlighted_bytes_str: str, highlighted_bytes: BytesMatch):
    """
    Find the position of bytes_str in surrounding_byte_str. Both args are raw text dumps of binary data.
    Because strings are longer than bytes (stuff like '\xcc' are 4 chars when printed are one byte and the ANSI unprintables
    include stuff like 'NegativeAcknowledgement' which is over 20 chars) they represent so we have to re-find the location to highlight the bytes
    correctly.
    """
    # Start a few chars in to avoid errors: sometimes we're searching for 1 or 2 bytes and there's a false positive
    # in the extra bytes. Tthis isn't perfect - it's starting us at the first index into the *bytes* that's safe to
    # check but this is almost certainly far too soon given the large % of bytes that take 4 chars to print ('\x02' etc)
    highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str, highlighted_bytes.highlight_start_idx)

    # TODO: Somehow \' and ' don't always come out the same :(
    if highlight_idx == -1:
        log.info(f"Failed to find highlighted_bytes in first pass so deleting single quotes and retrying." + \
                "  Highlighting may be off by a few chars,")

        surrounding_bytes_str = surrounding_bytes_str.replace("\\'", "'")
        highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str)

        if highlight_idx == -1:
            log.warning(f"Failed to find\n{highlighted_bytes_str}\nin surrounding bytes:\n{surrounding_bytes_str}")
            log.warning("Highlighting will not work on this decoded string.")

    return highlight_idx



"""
Constants related to character encodings
* https://www.mit.edu/people/kenta/two/iso8859.html
* https://www.utf8-chartable.de/unicode-utf8-table.pl?utf8=dec
"""

# Bytes (TODO: why is this here?)
NEWLINE_BYTE = b"\n"

# String constants
ENCODING = 'encoding'
ASCII = 'ascii'
UTF_8 = 'utf-8'
UTF_16 = 'utf-16'
UTF_32 = 'utf-32'
ISO_8859_1 =  'iso-8859-1'
WINDOWS_1252 = 'windows-1252'


# Byte order marks
BOMS = {
    b'\x2b\x2f\x76':     'UTF-7 BOM',
    b'\xef\xbb\xbf':     'UTF-8 BOM',
    b'\xfe\xff':         'UTF-16 BOM, big-endian',
    b'\xff\xfe':         'UTF-16 BOM, little-endian',
    b'\xff\xfe\x00\x00': 'UTF-32 BOM, little-endian',
    b'\x00\x00\xfe\xff': 'UTF-32 BOM, big-endian',
    b'\x0e\xfe\xff':     'SCSU BOM',
}


# ASCII characters that either print nothing, put the cursor in a weird place, or (worst of all) actively
# delete stuff you already printed
UNPRINTABLE_ASCII = {
    0: 'NUL',
    1: 'SOH',  # 'StartHeading',
    2: 'STX',  # 'StartText',
    3: 'ETX',
    4: 'EOT',  # End of transmission
    5: 'ENQ',  # 'Enquiry',
    6: 'ACK',  # 'Acknowledgement',
    7: 'BEL',  # 'Bell',
    8: 'BS',   # 'BackSpace',
    #9:  'HT'  # 'HorizontalTab',
    #10: 'LF',  # 'LineFeed',
    11: 'VT',  # 'VerticalTab',
    12: 'FF',  # 'FormFeed', AKA 'NewPage'
    13: 'CR',  # 'CarriageReturn',
    14: 'SO',  # 'ShiftOut',
    15: 'SI',  # 'ShiftIn',
    16: 'DLE',  # 'DataLineEscape',
    17: 'DC1',  # DeviceControl1',
    18: 'DC2',  # 'DeviceControl2',
    19: 'DC3',  # 'DeviceControl3',
    20: 'DC4',  # 'DeviceControl4',
    21: 'NAK',   # NegativeAcknowledgement',
    22: 'SYN',  # 'SynchronousIdle',
    23: 'ETB',  # 'EndTransmitBlock',
    24: 'CAN',  # 'Cancel',
    25: 'EM',  # 'EndMedium',
    26: 'SUB',  # 'Substitute',
    27: 'ESC',  # 'Escape',
    28: 'FS',  # 'FileSeparator',
    29: 'GS',  #'GroupSeparator',
    30: 'RS',  #'RecordSeparator',
    31: 'US',  # 'UnitSeparator',
    127: 'DEL', # Delete
}


# Fill in a dict with integer keys/values corresponding to where a given chr encoding has no chars bc
# this range is for C1 control chars. AKA # The 'undefined' part of the character map.
def scrub_c1_control_chars(char_map):
    for i in range(128, 160):
        char_map[i] = f"C1_{i}"


# ISO-8859-1 AKA "Latin-1". Basically ASCII but using more of 128-256    http://www.gammon.com.au/unicode/
UNPRINTABLE_ISO_8859_1 = UNPRINTABLE_ASCII.copy()
scrub_c1_control_chars(UNPRINTABLE_ISO_8859_1)

UNPRINTABLE_ISO_8859_1.update({
    129: 'HOP',
    141: 'RLF',
    160: 'NBSP',
    173: 'SHY',
})


# UTF-8 Makes no use of 128-256 on their own, only as continuation bytes.
# https://en.wikipedia.org/wiki/UTF-8
UNPRINTABLE_UTF_8 = UNPRINTABLE_ASCII.copy()
# The C1 bytes can appear but only as continuations
#scrub_c1_control_chars(UNPRINTABLE_UTF_8)

# C0, C1, FE, and FF, etc. *never* appear in UTF-8
UNPRINTABLE_UTF_8.update({
    192: 'C0',
    193: 'C1',
    245: 'F5',
    246: 'F6',
    247: 'F7',
    248: 'F8',
    249: 'F9',
    250: 'FA',
    251: 'FB',
    252: 'FC',
    253: 'FD',
    254: 'FE',
    255: 'FF',
})


# Win_1252 is a lot like other 256 char encodings but they colonized the C1 char DMZ in the middle
UNPRINTABLE_WIN_1252 = UNPRINTABLE_ASCII.copy()

UNPRINTABLE_WIN_1252.update({
    129: 'HOP', # High Octet Preset
    141: 'RLF', # Reverse Line Feed
    143: 'SS3', # Single shift 3
    144: 'DCS', # Device Control String
    147: 'STS', # Set transmit state
    160: 'NBSP',
})


# ISO-8859-7     #http://www.gammon.com.au/unicode/
UNPRINTABLE_ISO_8859_7 = UNPRINTABLE_ASCII.copy()
scrub_c1_control_chars(UNPRINTABLE_ISO_8859_7)

UNPRINTABLE_ISO_8859_7.update({
    174: 'AE',
    210: 'D2',
    255: 'FF'
})


# The encodings we will attempt to actually use
# Values are the unprintable values in that encoding in a dict (keys in dict are ints)
ENCODINGS_TO_ATTEMPT = {
    ASCII:        UNPRINTABLE_ASCII,
    UTF_8:          UNPRINTABLE_UTF_8,
    UTF_16:         None,
    UTF_32:         None,  # UTF-16 and 32 are handled differently
    #'utf-7':
    ISO_8859_1:   UNPRINTABLE_ISO_8859_1,
    WINDOWS_1252: UNPRINTABLE_WIN_1252
}


SINGLE_BYTE_ENCODINGS = [ASCII, ISO_8859_1, WINDOWS_1252]

# Unused cruft (mostly Asian language encodings)
ENCODINGS = [
    'big5',
    'big5hkscs',
    'cp950',
    'gb2312',
    'gbk',
    'gb18030',
    'hz',
    'iso2022_jp_2',
    'utf-7',
    'utf-8',
    'utf-16',
]



import re
from collections import namedtuple
from numbers import Number
from os import path
from typing import Any, Dict, List, Union
from pdfalyzer.binary.bytes_match import BytesMatch

import yara
from rich.padding import Padding
from rich.text import Text

from pdfalyzer.binary.bytes_decoder import BytesDecoder
from pdfalyzer.detection.constants.javascript_reserved_keywords import HIGH_VALUE_KEYWORDS
from pdfalyzer.helpers.bytes_helper import clean_byte_string
from pdfalyzer.helpers.file_helper import load_binary_data
from pdfalyzer.helpers.rich_text_helper import console, console_width, dim_if
from pdfalyzer.util.filesystem_awareness import PROJECT_DIR
from pdfalyzer.util.logging import log

# Some rules are applied to the whole file, others we only want to apply to decrypted/decoded binary streams
YARA_RULE_DIR = path.join(PROJECT_DIR, 'yara')
YARA_PDF_RULES_FILE = path.join(YARA_RULE_DIR, 'PDF.yara')
YARA_STREAM_RULES_FILE = path.join(YARA_RULE_DIR, 'PDF_binary_stream.yara')
YARA_PDF_RULES = yara.compile(YARA_PDF_RULES_FILE)
YARA_STREAM_RULES = yara.compile(YARA_STREAM_RULES_FILE)

URL_REGEX = re.compile('^https?:')
DIGITS_REGEX = re.compile("^\\d+$")
HEX_REGEX = re.compile('^[0-9A-Fa-f]+$')
DATE_REGEX = re.compile('\\d{4}-\\d{2}-\\d{2}')
MATCHER_VAR_REGEX = re.compile('\\$[a-z_]+')

YARA_STRING_STYLES: Dict[re.Pattern, str] = {
    URL_REGEX: 'yara.url',
    DIGITS_REGEX: 'yara.number',
    HEX_REGEX: 'yara.hex',
    DATE_REGEX: 'yara.date',
    MATCHER_VAR_REGEX: 'yara.match_var'
}

YaraResults = namedtuple('YaraResults', ['matches', 'non_matches'])


class YaraScanner:
    def __init__(self, _bytes: bytes, label: Union[str, Text]) -> None:
        self.bytes: bytes = _bytes
        self.label: Text = Text(label) if isinstance(label, str) else label
        self.matches = []
        self.non_matches = []

    @classmethod
    def for_file(cls, file_path: str) -> 'YaraScanner':
        """Load the binary data from file_path"""
        return cls(load_binary_data(file_path), path.basename(file_path))

    def scan(self) -> None:
        def yara_callback(data: dict):
            if data['matches']:
                self.matches.append(data)
            else:
                self.non_matches.append(data)

            return yara.CALLBACK_CONTINUE

        YARA_PDF_RULES.match(data=self.bytes, callback=yara_callback)
        YARA_STREAM_RULES.match(data=self.bytes, callback=yara_callback)
        self.any_matches = (len(self.matches) > 0)

        for yara_match in self.matches:
            console.print(self.label + Text(f" matched yara rule '{yara_match['rule']}'!", style='error'))
            console.print(rich_yara_match(yara_match))
            console.line()

            for match in BytesMatch.for_yara_strings_in_match(self.bytes, yara_match):
                BytesDecoder(match, yara_match['rule']).print_decode_attempts()

        non_matches_text = sorted([Text(nm['rule'], 'grey') for nm in self.non_matches], key=str)

        # Only show the non matches if there were valid ones, otherwise just show the number
        if self.any_matches:
            non_match_desc = f" did not match the other {len(self.non_matches)} yara rules"
            console.print(self.label + Text(non_match_desc, style='grey.dark') + Text(': '))
            console.print(Padding(Text(', ', 'white').join(non_matches_text), (0, 0, 1, 4)))
        else:
            non_match_desc = f" did not match any of the {len(self.non_matches)} yara rules"
            console.print(dim_if(self.label + Text(non_match_desc, style='grey.dark'), True))


def rich_yara_match(element: Any, depth: int = 0) -> Text:
    """Mildly painful/hacky way of coloring a yara result hash"""
    indent = Text((depth + 1) * 4 * ' ')
    end_indent = Text(depth * 4 * ' ')

    if isinstance(element, str):
        txt = yara_string(element)
    elif isinstance(element, bytes):
        txt = Text(clean_byte_string(element), style='bytes')
    elif isinstance(element, Number):
        txt = Text(str(element), style='bright_cyan')
    elif isinstance(element, bool):
        txt = Text(str(element), style='red' if not element else 'green')
    elif isinstance(element, (list, tuple)):
        if len(element) == 0:
            txt = Text('[]', style='white')
        else:
            total_length = sum([len(str(e)) for e in element]) + ((len(element) - 1) * 2) + + len(indent) + 2
            elements_txt = [rich_yara_match(e, depth + 1) for e in element]
            list_txt = Text('[', style='white')

            if total_length > console_width() or len(element) > 3:
                join_txt = Text(f"\n{indent}" )
                list_txt.append(join_txt).append(Text(f",{join_txt}").join(elements_txt))
                list_txt += Text(f'\n{end_indent}]', style='white')
            else:
                list_txt += Text(', ').join(elements_txt) + Text(']')

            return list_txt
    elif isinstance(element, dict):
        element = {k: v for k, v in element.items() if k not in ['matches', 'rule']}

        if len(element) == 0:
            return Text('{}')

        txt = Text('{\n', style='white')

        for i, k in enumerate(element.keys()):
            v = element[k]
            txt += indent + Text(f"{k}: ", style='yara.key') + rich_yara_match(v, depth + 1)

            if (i + 1) < len(element.keys()):
                txt.append(",\n")
            else:
                txt.append("\n")

        txt += end_indent + Text('}', style='white')
    else:
        log.warn(f"Unknown yara return of type {type(element)}: {element}")
        txt = indent + Text(str(element))

    return txt


def yara_string(_string: str) -> Text:
    for regex in YARA_STRING_STYLES.keys():
        if regex.match(_string):
            return Text(_string, YARA_STRING_STYLES[regex])

    return Text(_string, style='yara.string')



from numbers import Number
from typing import Union

from rich.text import Text

from yaralyzer.helpers.rich_text_helper import prefix_with_plain_text_obj


def is_divisible_by(n: int, divisor: int) -> bool:
    return divmod(n, divisor)[1] == 0


def is_even(n: int) -> bool:
    return is_divisible_by(n, 2)



"""
Walks the PDF objects, wrapping each in a PdfTreeNode and putting them into a tree
managed by the anytree library. Once the PDF is parsed this class manages things like
searching the tree and printing out information.
"""
from collections import defaultdict
from os.path import basename
from typing import List, Optional

from anytree import LevelOrderIter, RenderTree, SymlinkNode
from anytree.render import DoubleStyle
from anytree.search import findall, findall_by_attr
from PyPDF2 import PdfReader
from PyPDF2.errors import PdfReadError
from PyPDF2.generic import IndirectObject, NameObject, NumberObject, StreamObject
from rich.markup import escape
from rich.panel import Panel
from rich.table import Column, Table
from rich.text import Text
from yaralyzer.config import YaralyzerConfig
from yaralyzer.helpers.bytes_helper import get_bytes_info
from yaralyzer.helpers.file_helper import load_binary_data
from yaralyzer.helpers.rich_text_helper import CENTER, LEFT, size_in_bytes_text
from yaralyzer.output.rich_console import BYTES_HIGHLIGHT, GREY, console
from yaralyzer.output.rich_layout_elements import bytes_hashes_table
from yaralyzer.util.logging import log

from pdfalyzer.binary.binary_scanner import BinaryScanner
from pdfalyzer.decorators.document_model_printer import print_with_header
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.detection.yaralyzer_helper import get_file_yaralyzer
from pdfalyzer.helpers.pdf_object_helper import INDETERMINATE_REFERENCES, get_symlink_representation
from pdfalyzer.helpers.string_helper import pp
from pdfalyzer.font_info import FontInfo
from pdfalyzer.output.layout import print_section_header, print_section_subheader, print_section_sub_subheader
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

TRAILER_FALLBACK_ID = 10000000


class Pdfalyzer:
    def __init__(self, pdf_path: str):
        self.pdf_path = pdf_path
        self.pdf_basename = basename(pdf_path)
        self.pdf_bytes = load_binary_data(pdf_path)
        self.pdf_bytes_info = get_bytes_info(self.pdf_bytes)
        pdf_file = open(pdf_path, 'rb')  # Filehandle must be left open for PyPDF2 to perform seeks
        self.pdf = PdfReader(pdf_file)
        self.yaralyzer = get_file_yaralyzer(pdf_path)
        # Initialize tracking variables
        self.indeterminate_ids = set()  # See INDETERMINATE_REFERENCES comment
        self.traversed_nodes = {}
        self.font_infos = []
        self.max_generation = 0
        # Build the tree
        self.walk_pdf()

    def walk_pdf(self):
        """
        PDFs are read trailer first so trailer is the root of the tree.
        We build the rest by recursively following references we find in nodes we encounter.
        """
        trailer = self.pdf.trailer
        self.pdf_size = trailer.get(SIZE)

        # Technically the trailer has no ID in the PDF but we set it to the /Size of the PDF for convenience
        if self.pdf_size is None:
            trailer_id = TRAILER_FALLBACK_ID
        else:
            trailer_id = self.pdf_size

        self.pdf_tree = PdfTreeNode(trailer, TRAILER, trailer_id)
        self.traversed_nodes[self.pdf_tree.idnum] = self.pdf_tree
        self.walk_node(self.pdf_tree)
        self._resolve_indeterminate_nodes()  # After scanning all the objects we place nodes whose position was uncertain
        self._extract_font_infos()
        self._verify_all_traversed_nodes_are_in_tree()
        self._verify_untraversed_nodes_are_untraversable()
        self._symlink_other_relationships()  # Create symlinks for non parent/child relationships between nodes
        log.info(f"Walk complete.")

    def walk_node(self, node: PdfTreeNode):
        """Recursively walk the PDF's tree structure starting at a given node"""
        log.info(f'walk_node() called with {node}. Object dump:\n{print_with_header(node.obj, node.label)}')
        self._ensure_safe_to_walk(node)
        references = node.references()

        for (key, address, value) in references:
            self._process_reference(node, key, address, value)

        node.all_references_processed = True

        for next_node in [self.traversed_nodes[ref.pdf_obj.idnum] for ref in references]:
            if not next_node.all_references_processed:
                self.walk_node(next_node)

    def find_node_by_idnum(self, idnum):
        """Find node with idnum in the tree. Return None if that node is not in tree or not reachable from the root."""
        nodes = [n for n in findall_by_attr(self.pdf_tree, name='idnum', value=idnum) if not isinstance(n, SymlinkNode)]

        if len(nodes) == 0:
            return None
        elif len(nodes) == 1:
            return nodes[0]
        else:
            raise PdfWalkError(f"Too many nodes had id {idnum}: {nodes}")

    def print_everything(self) -> None:
        """Print every kind of analysis on offer to Rich console"""
        self.print_document_info()
        self.print_summary()
        self.print_tree()
        self.print_rich_table_tree()
        self.print_font_info()
        self.print_other_relationships()

    def print_document_info(self) -> None:
        """Print the embedded document info (author, timestamps, version, etc)"""
        print_section_header(f'Document Info for {self.pdf_basename}')
        console.print(pp.pformat(self.pdf.getDocumentInfo()))
        console.line()
        console.print(bytes_hashes_table(self.pdf_bytes, self.pdf_basename))
        console.line()

    def print_tree(self):
        print_section_header(f'Simple tree view of {self.pdf_basename}')

        for pre, _fill, node in RenderTree(self.pdf_tree, style=DoubleStyle):
            if isinstance(node, SymlinkNode):
                symlink_rep = get_symlink_representation(node.parent, node)
                console.print(pre + f"[{symlink_rep.style}]{symlink_rep.text}[/{symlink_rep.style}]")
            else:
                console.print(Text(pre) + node.__rich__())

        self._verify_all_traversed_nodes_are_in_tree()
        console.print("\n\n")

    def print_rich_table_tree(self) -> None:
        print_section_header(f'Rich tree view of {self.pdf_basename}')
        console.print(self.pdf_tree.generate_rich_tree())
        self._verify_all_traversed_nodes_are_in_tree()

    def print_summary(self) -> None:
        print_section_header(f'PDF Node Summary for {self.pdf_basename}')
        console.print_json(data=self._analyze_tree(), sort_keys=True)

    def print_font_info(self, font_idnum=None) -> None:
        print_section_header(f'{len(self.font_infos)} fonts found in {self.pdf_basename}')

        for font_info in [fi for fi in self.font_infos if font_idnum is None or font_idnum == fi.idnum]:
            font_info.print_summary()

    def print_streams_analysis(self, idnum: Optional[int] = None) -> None:
        print_section_header(f'Binary Stream Analysis / Extraction')
        console.print(self.stream_objects_table())

        for node in [n for n in self.stream_nodes() if idnum is None or idnum == n.idnum]:
            node_stream_bytes = node.stream_data

            if node_stream_bytes is None or node.stream_length == 0:
                print_section_sub_subheader(f"{node} stream has length 0", style='dim')
                continue

            if not isinstance(node_stream_bytes, bytes):
                msg = f"Stream in {node} is not bytes, it's {type(node.stream_data)}. Will reencode for YARA " + \
                       "but they may not be the same bytes as the original stream!"
                log.warning(msg)
                node_stream_bytes = node_stream_bytes.encode()

            print_section_subheader(f"{escape(str(node))} Summary and Analysis", style=f"{BYTES_HIGHLIGHT} reverse")
            binary_scanner = BinaryScanner(node_stream_bytes, node)
            console.print(bytes_hashes_table(binary_scanner.bytes))
            binary_scanner.print_stream_preview()
            binary_scanner.check_for_dangerous_instructions()

            if not YaralyzerConfig.SUPPRESS_DECODES:
                binary_scanner.check_for_boms()
                binary_scanner.force_decode_all_quoted_bytes()

            binary_scanner.print_decoding_stats_table()

    def print_yara_results(self, font_idnum=None) -> None:
        print_section_header(f"YARA Scan of PDF rules for '{self.pdf_basename}'")
        self.yaralyzer.yaralyze()

        # TODO: we should really scan all the binary streams not just those in the fonts
        for font_info in [fi for fi in self.font_infos if font_idnum is None or font_idnum == fi.idnum]:
            font_info.yara_scan()

    def print_other_relationships(self) -> None:
        """Print the inter-node, non-tree relationships for all nodes in the tree"""
        console.print("\n\n")
        console.print(Panel(f"Other Relationships", expand=False), style='reverse')

        for node in LevelOrderIter(self.pdf_tree):
            if len(node.other_relationships) == 0:
                continue

            console.print("\n")
            console.print(Panel(f"Non tree relationships for {node}", expand=False))
            node.print_other_relationships()

    def stream_objects_table(self) -> Table:
        table = Table('Stream Length', 'Node')
        table.columns[0].justify = 'right'

        for node in self.stream_nodes():
            table.add_row(size_in_bytes_text(node.stream_length), node.__rich__())

        return table

    def stream_nodes(self) -> List[PdfTreeNode]:
        """List of actual nodes (not SymlinkNodes) containing streams sorted by PDF object ID"""
        stream_filter = lambda node: node.contains_stream() and not isinstance(node, SymlinkNode)
        return sorted(findall(self.pdf_tree, stream_filter), key=lambda r: r.idnum)

    def _process_reference(self, node: PdfTreeNode, key: str, address: str, reference: IndirectObject) -> [PdfTreeNode]:
        """
        Place the referenced node in the tree. Returns a list of nodes to walk next.
        address is the reference_key used in node.obj to refer to 'reference' object
           plus any modifiers like [2] or [/Something]
        """
        self.max_generation = max([self.max_generation, reference.generation or 0])
        seen_before = (reference.idnum in self.traversed_nodes)
        referenced_node = self._build_or_find_node(reference, address)
        reference_log_string = f"{node} reference at {address} to {referenced_node}"
        log.info(f'Assessing {reference_log_string}...')
        references_to_return = []
        is_pure_ref = node.is_pure_reference(key)
        log.debug(f"   {node} => {key}: is_pure_ref? {is_pure_ref}")

        # If one is already a parent/child of the other there's nothing to do
        if referenced_node == node.parent or referenced_node in node.children:
            log.debug(f"  {node} and {referenced_node} are already related")
            return []

        # If there's an explicit /Parent or /Kids reference then we know the correct relationship
        if node.is_parent_reference(key) or node.is_child_reference(key):
            log.debug(f"Explicit parent/child reference in {node} at {key}")
            if node.is_parent_reference(key):
                # try:
                    node.set_parent(referenced_node)
                # except Exception as e:
                #     import pdb;pdb.set_trace()
            else:
                node.add_child(referenced_node)

            if reference.idnum in self.indeterminate_ids:
                log.info(f"  Found refefence {address} => {node} of previously indeterminate node {referenced_node}")
                self.indeterminate_ids.remove(reference.idnum)

            if not seen_before:
                references_to_return = [referenced_node]

        # Indeterminate references need to wait until everything has been scanned to be placed
        elif node.is_indeterminate_reference(key):
            log.info(f'  Indeterminate {reference_log_string}')
            referenced_node.add_relationship(node, address)
            self.indeterminate_ids.add(referenced_node.idnum)
            return [referenced_node]

        # Non tree references are not children or parents.
        elif node.is_pure_reference(key):
            log.debug(f"{reference_log_string} is a pure reference.")
            referenced_node.add_relationship(node, address)

            # If node looks like a pure ref but is not in tree, consider it indeterminate
            if not self.find_node_by_idnum(referenced_node.idnum):
                references_to_return = [referenced_node]
                self.indeterminate_ids.add(referenced_node.idnum)

        # If we've seen the node before it should have a parent or be indeterminate
        elif seen_before:
            if reference.idnum not in self.indeterminate_ids and referenced_node.parent is None:
                raise PdfWalkError(f"{reference_log_string} - ref has no parent and is not indeterminate")

            referenced_node.add_relationship(node, address)

        # If no other conditions are met, add the reference as a child
        else:
            node.add_child(referenced_node)
            references_to_return = [referenced_node]

        log.debug("Nodes to walk next: " + ', '.join([str(r) for r in references_to_return]))
        return references_to_return

    def _symlink_other_relationships(self):
        """Create SymlinkNodes for relationships between PDF objects that are not parent/child relationships"""
        for node in LevelOrderIter(self.pdf_tree):
            if node.other_relationship_count() == 0 or isinstance(node, SymlinkNode):
                continue

            log.info(f"Symlinking {node}'s {node.other_relationship_count()} other relationships...")

            for relationship in node.other_relationships:
                log.debug(f"   * Linking {relationship}")
                SymlinkNode(node, parent=relationship.from_node)

    def _resolve_indeterminate_nodes(self) -> None:
        """
        Some nodes cannot be placed until we have walked the rest of the tree. For instance
        if we encounter a /Page that references /Resources we need to know if there's a
        /Pages parent of the /Page before committing to a tree structure.
        """
        indeterminate_nodes = [self.traversed_nodes[idnum] for idnum in self.indeterminate_ids]
        indeterminate_nodes_string = "\n   ".join([f"{node}" for node in indeterminate_nodes])
        log.info(f"Resolving {len(indeterminate_nodes)} indeterminate nodes: {indeterminate_nodes_string}")

        for idnum in self.indeterminate_ids:
            if self.find_node_by_idnum(idnum):
                log.warning(f"Node with ID {idnum} marked indeterminate but found in tree...")
                continue

            set_lowest_id_node_as_parent = False
            node = self.traversed_nodes[idnum]
            referenced_by_keys = list(set([r.reference_key for r in node.other_relationships]))
            log.info(f"Attempting to resolve indeterminate node {node}")

            if node.parent is not None:
                log.info(f"{node} already has parent: {node.parent}")
                continue

            if node.label == RESOURCES:
                self._place_resources_node(node)
                continue
            # TODO: these almost all have the same outcome; could be one super ugly if statement
            elif node.label == COLOR_SPACE:
                log.info("Color space node found; placing at lowest ID")
                set_lowest_id_node_as_parent = True
                possible_parents = node.other_relationships
            elif len(referenced_by_keys) == 1:
                log.info(f"{node}'s other relationships all use key {referenced_by_keys[0]}, linking to lowest id")
                set_lowest_id_node_as_parent = True
                possible_parents = node.other_relationships
            elif all([EXTERNAL_GRAPHICS_STATE_REGEX.match(key) for key in referenced_by_keys]):
                log.info(f"{node}'s other relationships are all {EXT_G_STATE} refs; linking to lowest id")
                set_lowest_id_node_as_parent = True
                possible_parents = node.other_relationships
            elif len(referenced_by_keys) == 2 and \
                    (   referenced_by_keys[0] in referenced_by_keys[1] \
                     or referenced_by_keys[1] in referenced_by_keys[0]):
                log.info(f"{node}'s other relationships ref keys are same except slice: {referenced_by_keys}, linking to lowest id")
                set_lowest_id_node_as_parent = True
                possible_parents = node.other_relationships
            elif any(r.from_node.label == RESOURCES for r in node.other_relationships) and \
                    all(any(r.from_node.label.startswith(ir) for ir in INDETERMINATE_REFERENCES) for r in node.other_relationships):
                log.info(f"Linking {node} to lowest id {RESOURCES} node...")
                possible_parents = [r for r in node.other_relationships if r.from_node.label == RESOURCES]
                set_lowest_id_node_as_parent = True
            else:
                determinate_relations = [r for r in node.other_relationships if r.from_node.label not in INDETERMINATE_REFERENCES]
                determinate_refkeys = set([r.from_node.label for r in determinate_relations])

                if len(determinate_refkeys) == 1:
                    ref_key = determinate_relations[0].reference_key
                    log.info(f"Only one ref key {ref_key} that's determinate, choose parent as lowest id using it")
                    possible_parents = determinate_relations
                    set_lowest_id_node_as_parent = True

            if set_lowest_id_node_as_parent:
                lowest_idnum = min([r.from_node.idnum for r in possible_parents])
                lowest_id_relationship = next(r for r in node.other_relationships if r.from_node.idnum == lowest_idnum)
                log.info(f"Setting parent of {node} to {lowest_id_relationship}")
                node.set_parent(self.traversed_nodes[lowest_idnum])
                continue

            self.print_tree()
            node.print_other_relationships()
            log.fatal("Dumped tree status and other_relationships for debugging")
            raise PdfWalkError(f"Cannot place {node}")

    def _extract_font_infos(self) -> None:
        """Extract information about fonts in the tree and place it in self.font_infos"""
        for node in LevelOrderIter(self.pdf_tree):
            if isinstance(node.obj, dict) and RESOURCES in node.obj:
                log.debug(f"Extracting fonts from node with '{RESOURCES}' key: {node}...")
                known_font_ids = [fi.idnum for fi in self.font_infos]

                self.font_infos += [
                    fi for fi in FontInfo.extract_font_infos(node.obj)
                    if fi.idnum not in known_font_ids
                ]

    def _ensure_safe_to_walk(self, node) -> None:
        if not node.idnum in self.traversed_nodes:
            return

        if self.traversed_nodes[node.idnum] != node:
            raise PdfWalkError("Duplicate PDF object ID {node.idnum}")

    def _place_resources_node(self, resources_node) -> None:
        """See if there is a common ancestor like /Pages; if so that's the parent"""
        relationships_labels = set()

        for relationship in resources_node.other_relationships:
            other_relationships = [r for r in resources_node.other_relationships if r != relationship]
            relationships_labels.add(relationship.from_node.label)

            if all(relationship[0] in r[0].ancestors for r in other_relationships):
                log.info(f'{relationship[0]} is the common ancestor found while placing /Resources')
                resources_node.set_parent(relationship[0])
                return

        if relationships_labels == set([PAGE, PAGES]):
            pages_nodes = [r.from_node for r in resources_node.other_relationships if r.from_node.label == PAGES]
            log.warning(f"Failed to place {resources_node}; seems to be a loose {PAGE}. Linking to first {PAGES}")
            resources_node.set_parent(sorted(pages_nodes, key=lambda n: n.idnum)[0])
            return

        log.error(f"Failed to place {resources_node}. {RESOURCES} relationship dump:")
        resources_node.print_other_relationships()
        raise PdfWalkError(f'Failed to place {resources_node}')

    def _build_or_find_node(self, reference: IndirectObject, reference_key: str) -> PdfTreeNode:
        """If node exists in self.traversed_nodes return it, otherwise build a node"""
        if reference.idnum in self.traversed_nodes:
            return self.traversed_nodes[reference.idnum]

        # TODO: known_to_parent_as should not be passed for non-child relationships even though as it
        #       stands it is corrected later when the true parent is found.
        log.debug(f"Building node for {reference_key} -> {reference}")
        new_node = PdfTreeNode.from_reference(reference, reference_key)
        self.traversed_nodes[reference.idnum] = new_node
        return new_node

    def _analyze_tree(self) -> dict:
        """Generate a dict with some basic data points about the PDF tree"""
        pdf_object_types = defaultdict(int)
        node_labels = defaultdict(int)
        keys_encountered = defaultdict(int)
        node_count = 0

        for node in LevelOrderIter(self.pdf_tree):
            pdf_object_types[type(node.obj).__name__] += 1
            node_labels[node.label] += 1
            node_count += 1

            if isinstance(node.obj, dict):
                for k in node.obj.keys():
                    keys_encountered[k] += 1

        return {
            'keys_encountered': keys_encountered,
            'node_count': node_count,
            'node_labels': node_labels,
            'pdf_object_types': pdf_object_types,
        }

    def _print_traversed_nodes(self) -> None:
        """Debug method that displays which nodes have already been walked"""
        for i in sorted(self.traversed_nodes.keys()):
            console.print(f'{i}: {self.traversed_nodes[i]}')

    def _verify_all_traversed_nodes_are_in_tree(self) -> None:
        """Make sure every node we can see is reachable from the root of the tree"""
        missing_nodes = [
            node for idnum, node in self.traversed_nodes.items()
            if self.find_node_by_idnum(idnum) is None
        ]

        if len(missing_nodes) > 0:
            msg = f"Nodes were traversed but never placed: {escape(str(missing_nodes))}"
            console.print(msg)
            log.warning(msg)
            #raise PdfWalkError(msg)

    def _verify_untraversed_nodes_are_untraversable(self) -> None:
        """Make sure any PDF object IDs we can't find in tree are /ObjStm or /Xref nodes"""
        if self.pdf_size is None:
            log.warning(f"{SIZE} not found in PDF trailer; cannot verify all nodes are in tree")
            return

        if self.max_generation > 0:
            log.warning(f"_verify_untraversed_nodes_are_untraversable() only checking generation {self.max_generation}")

        for idnum in [i + 1 for i in range(self.pdf_size)]:
            if self.find_node_by_idnum(idnum) is not None:
                log.debug(f"Verified object {idnum} is in tree")
                continue

            ref = IndirectObject(idnum, self.max_generation, self.pdf)
            try:
                obj = ref.get_object()
            except PdfReadError as e:
                if 'Invalid Elementary Object' in str(e):
                    log.info(f"Couldn't verify elementary obj with id {idnum} is properly in tree")
                    continue
                else:
                    log.error(str(e))
                    console.print(str(e), style='error')
                    console.print_exception()
                    raise

            if obj is None:
                log.error(f"Cannot find ref {ref} in PDF!")
                continue
            elif isinstance(obj, (NumberObject, NameObject)):
                log.info(f"Obj {idnum} is a {type(obj)} w/value {obj}; if referenced by /Length etc. this is a nonissue but maybe worth doublechecking")
                continue
            elif not isinstance(obj, dict):
                log.error(f"Obj {idnum} ({obj}) of type {type(obj)} isn't dict, cannot determine if it should be in tree")
                continue
            elif TYPE not in obj:
                msg = f"Obj {idnum} has no {TYPE} and is not in tree. Either a loose node w/no data or an error in pdfalyzer."
                msg += f"\nHere's the contents for you to assess:\n{obj}"
                log.warning(msg)
                continue

            obj_type = obj[TYPE]

            if obj_type == OBJECT_STREAM:
                log.debug(f"Object with id {idnum} not found in tree because it's an {OBJECT_STREAM}")
            elif obj[TYPE] == XREF:
                placeable = XREF_STREAM in self.pdf.trailer

                for k, v in self.pdf.trailer.items():
                    xref_val_for_key = obj.get(k)

                    if k in [XREF_STREAM, PREV]:
                        continue
                    elif k == SIZE:
                        if xref_val_for_key is None or v != (xref_val_for_key + 1):
                            log.info(f"{XREF} has {SIZE} of {xref_val_for_key}, trailer has {SIZE} of {v}")
                            placeable = False

                        continue
                    elif k not in obj or v != obj.get(k):
                        log.info(f"Trailer has {k} -> {v} but {XREF} obj has {obj.get(k)} at that key")
                        placeable = False

                if placeable:
                    self.pdf_tree.add_child(self._build_or_find_node(ref, XREF_STREAM))
            else:
                log.warning(f"{XREF} Obj {idnum} not found in tree!")



"""
Class for handling binary data - scanning through it for various suspicious patterns as well as forcing
various character encodings upon it to see what comes out.
"""
from collections import defaultdict
from numbers import Number
from typing import Any, Iterator, Optional, Pattern, Tuple, Union

from deprecated import deprecated
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from yaralyzer.bytes_match import BytesMatch
from yaralyzer.config import YaralyzerConfig
from yaralyzer.decoding.bytes_decoder import BytesDecoder
from yaralyzer.encoding_detection.character_encodings import BOMS
from yaralyzer.helpers.bytes_helper import hex_string, print_bytes
from yaralyzer.helpers.rich_text_helper import CENTER, na_txt, prefix_with_plain_text_obj
from yaralyzer.helpers.string_helper import escape_yara_pattern
from yaralyzer.output.rich_console import BYTES_NO_DIM, console, console_width
from yaralyzer.output.regex_match_metrics import RegexMatchMetrics
from yaralyzer.yara.yara_rule_builder import HEX, REGEX, safe_label
from yaralyzer.yaralyzer import Yaralyzer
from yaralyzer.util.logging import log

from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.decorators.pdf_tree_node import PdfTreeNode
from pdfalyzer.detection.constants.binary_regexes import (BACKTICK, DANGEROUS_STRINGS, FRONTSLASH, GUILLEMET,
     QUOTE_PATTERNS)
from pdfalyzer.helpers.rich_text_helper import NOT_FOUND_MSG, generate_subtable, pad_header
from pdfalyzer.helpers.string_helper import generate_hyphen_line
from pdfalyzer.output.layout import half_width, print_headline_panel, print_section_sub_subheader
from pdfalyzer.util.adobe_strings import CONTENTS, CURRENTFILE_EEXEC

# For rainbow colors
CHAR_ENCODING_1ST_COLOR_NUMBER = 203


class BinaryScanner:
    def __init__(self, _bytes: bytes, owner: Union['FontInfo', 'PdfTreeNode'], label: Optional[Text] = None):
        """owner is an optional link back to the object containing this binary"""
        self.bytes = _bytes
        self.label = label
        self.owner = owner

        if label is None and isinstance(owner, PdfTreeNode):
             self.label = owner.__rich__()

        self.stream_length = len(_bytes)
        self.regex_extraction_stats = defaultdict(lambda: RegexMatchMetrics())
        self.suppression_notice_queue = []

    def check_for_dangerous_instructions(self) -> None:
        """Scan for all the strings in DANGEROUS_INSTRUCTIONS list and decode bytes around them"""
        print_section_sub_subheader("Scanning Binary For Anything 'Mad Sus'...", style=f"bright_red")

        for instruction in DANGEROUS_STRINGS:
            yaralyzer = self._pattern_yaralyzer(instruction, REGEX)
            yaralyzer.highlight_style = 'bright_red bold'
            self.process_yara_matches(yaralyzer, instruction, force=True)

    def check_for_boms(self) -> None:
        """Check the binary data for BOMs"""
        print_section_sub_subheader("Scanning Binary for any BOMs...", style='BOM')

        for bom_bytes, bom_name in BOMS.items():
            yaralyzer = self._pattern_yaralyzer(hex_string(bom_bytes), HEX, bom_name)
            yaralyzer.highlight_style = 'BOM'
            self.process_yara_matches(yaralyzer, bom_name, force=True)

    def force_decode_all_quoted_bytes(self) -> None:
        """Find all strings matching QUOTE_PATTERNS (AKA between quote chars) and decode them with various encodings"""
        quote_types = QUOTE_PATTERNS.keys() if PdfalyzerConfig.QUOTE_TYPE is None else [PdfalyzerConfig.QUOTE_TYPE]

        for quote_type in quote_types:
            if self.owner and self.owner.type == CONTENTS and quote_type in [FRONTSLASH, GUILLEMET]:
                msg = f"Not attempting {quote_type} decode for {CONTENTS} node type..."
                print_headline_panel(msg, style='dim')
                continue

            quote_pattern = QUOTE_PATTERNS[quote_type]
            print_section_sub_subheader(f"Forcing Decode of {quote_type.capitalize()} Quoted Strings", style=BYTES_NO_DIM)
            yaralyzer = self._quote_yaralyzer(quote_pattern, quote_type)
            self.process_yara_matches(yaralyzer, f"{quote_type}_quoted")

    # -------------------------------------------------------------------------------
    # These extraction iterators will iterate over all matches for a specific pattern.
    # extract_regex_capture_bytes() is the generalized method that acccepts any regex.
    # -------------------------------------------------------------------------------
    def extract_guillemet_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Iterate on all strings surrounded by Guillemet quotes, e.g. «string»"""
        return self._quote_yaralyzer(QUOTE_PATTERNS[GUILLEMET], GUILLEMET).match_iterator()

    def extract_backtick_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Returns an interator over all strings surrounded by backticks"""
        return self._quote_yaralyzer(QUOTE_PATTERNS[BACKTICK], BACKTICK).match_iterator()

    def extract_front_slash_quoted_bytes(self) -> Iterator[Tuple[BytesMatch, BytesDecoder]]:
        """Returns an interator over all strings surrounded by front_slashes (hint: regular expressions)"""
        return self._quote_yaralyzer(QUOTE_PATTERNS[FRONTSLASH], FRONTSLASH).match_iterator()

    def print_stream_preview(self, num_bytes=None, title_suffix=None) -> None:
        """Print a preview showing the beginning and end of the stream data"""
        num_bytes = num_bytes or console_width()
        snipped_byte_count = self.stream_length - (num_bytes * 2)
        console.line()

        if snipped_byte_count < 0:
            title = f"ALL {self.stream_length} BYTES IN STREAM"
        else:
            title = f"FIRST AND LAST {num_bytes} BYTES OF {self.stream_length} BYTE STREAM"

        title += title_suffix if title_suffix is not None else ''
        title = f"BEGIN {title}".upper()
        console.print(generate_hyphen_line(title=title), style='dim')

        if snipped_byte_count < 0:
            print_bytes(self.bytes)
        else:
            print_bytes(self.bytes[:num_bytes])
            console.print(f"\n    <...skip {snipped_byte_count} bytes...>\n", style='dim')
            print_bytes(self.bytes[-num_bytes:])

        console.print(generate_hyphen_line(title=title), style='dim')
        console.line()

    def print_decoding_stats_table(self) -> None:
        """Diplay aggregate results on the decoding attempts we made on subsets of self.bytes"""
        stats_table = new_decoding_stats_table(self.label.plain if self.label else '')
        regexes_not_found_in_stream = []

        for matcher, stats in self.regex_extraction_stats.items():
            # Set aside the regexes we didn't find so that the ones we did find are at the top of the table
            if stats.match_count == 0:
                regexes_not_found_in_stream.append([str(matcher), NOT_FOUND_MSG, na_txt()])
                continue

            regex_subtable = generate_subtable(cols=['Metric', 'Value'])
            decodes_subtable = generate_subtable(cols=['Encoding', 'Decoded', 'Forced', 'Failed'])

            for metric, measure in vars(stats).items():
                if isinstance(measure, Number):
                    regex_subtable.add_row(metric, str(measure))

            for i, (encoding, count) in enumerate(stats.was_match_decodable.items()):
                decodes_subtable.add_row(
                    Text(encoding, style=f"color({CHAR_ENCODING_1ST_COLOR_NUMBER + 2 * i})"),
                    str(count),
                    str(self.regex_extraction_stats[matcher].was_match_force_decoded[encoding]),
                    str(self.regex_extraction_stats[matcher].was_match_undecodable[encoding]))

            stats_table.add_row(str(matcher), regex_subtable, decodes_subtable)

        for row in regexes_not_found_in_stream:
            row[0] = Text(row[0], style='color(235)')
            stats_table.add_row(*row, style='color(232)')

        console.line(2)
        console.print(stats_table, justify='center')

    def process_yara_matches(self, yaralyzer: Yaralyzer, pattern: str, force: bool = False) -> None:
        """Decide whether to attempt to decode the matched bytes, track stats. force param ignores min/max length"""
        for bytes_match, bytes_decoder in yaralyzer.match_iterator():
            self.regex_extraction_stats[pattern].match_count += 1
            self.regex_extraction_stats[pattern].bytes_matched += bytes_match.match_length
            self.regex_extraction_stats[pattern].bytes_match_objs.append(bytes_match)

            # Send suppressed decodes to a queue and track the reason for the suppression in the stats
            if not ((YaralyzerConfig.MIN_DECODE_LENGTH < bytes_match.match_length < YaralyzerConfig.MAX_DECODE_LENGTH) \
                    or force):
                self._queue_suppression_notice(bytes_match, pattern)
                continue

            # Print out any queued suppressed notices before printing non suppressed matches
            self._print_suppression_notices()
            self._record_decode_stats(bytes_match, bytes_decoder, pattern)

        # This check initializes the defaultdic for 'pattern'
        if self.regex_extraction_stats[pattern].match_count == 0:
            #console.print(f"{pattern} was not found for {escape(self.label.plain)}...", style='dim')
            pass

    def bytes_after_eexec_statement(self) -> bytes:
        """Get the bytes after the 'eexec' demarcation line (if it appears). See Adobe docs for details."""
        return self.bytes.split(CURRENTFILE_EEXEC)[1] if CURRENTFILE_EEXEC in self.bytes else self.bytes

    @deprecated(version='1.7.0', reason="Use process_yara_matches() instead")
    def extract_regex_capture_bytes(self, regex: Pattern[bytes]) -> Iterator[BytesMatch]:
        """Finds all matches of regex_with_one_capture in self.bytes and calls yield() with BytesMatch tuples"""
        for i, match in enumerate(regex.finditer(self.bytes, self._eexec_idx())):
            yield(BytesMatch.from_regex_match(self.bytes, match, i + 1))

    def _pattern_yaralyzer(
            self,
            pattern: str,
            pattern_type: str,
            rules_label: Optional[str] = None,
            pattern_label: Optional[str] = None
        ) -> Yaralyzer:
        """Build a yaralyzer to scan self.bytes"""
        return Yaralyzer.for_patterns(
            patterns=[escape_yara_pattern(pattern)],
            patterns_type=pattern_type,
            scannable=self.bytes,
            scannable_label=self.label.plain,
            rules_label=safe_label(rules_label or pattern),
            pattern_label=safe_label(pattern_label or pattern)
        )

    def _quote_yaralyzer(self, quote_pattern: str, quote_type: str):
        """Helper method to build a Yaralyzer for a quote_pattern"""
        label = f"{quote_type}_Quoted"

        if quote_type == GUILLEMET:
            return self._pattern_yaralyzer(quote_pattern, HEX, label, label)
        else:
            return self._pattern_yaralyzer(quote_pattern, REGEX, label, label)

    def _record_decode_stats(self, bytes_match: BytesMatch, decoder: BytesDecoder, label: str) -> None:
        """Attempt to decode _bytes with all configured encodings and print a table of the results"""
        # Track stats on whether the bytes were decodable or not w/a given encoding
        self.regex_extraction_stats[bytes_match.label].matches_decoded += 1

        for encoding, count in decoder.was_match_decodable.items():
            decode_stats = self.regex_extraction_stats[bytes_match.label].was_match_decodable
            decode_stats[encoding] = decode_stats.get(encoding, 0) + count

        for encoding, count in decoder.was_match_undecodable.items():
            failure_stats = self.regex_extraction_stats[bytes_match.label].was_match_undecodable
            failure_stats[encoding] = failure_stats.get(encoding, 0) + count

        for encoding, count in decoder.was_match_force_decoded.items():
            forced_stats = self.regex_extraction_stats[bytes_match.label].was_match_force_decoded
            forced_stats[encoding] = forced_stats.get(encoding, 0) + count

    def _queue_suppression_notice(self, bytes_match: BytesMatch, quote_type: str) -> None:
        """Print a message indicating that we are not going to decode a given block of bytes"""
        self.regex_extraction_stats[bytes_match.label].skipped_matches_lengths[bytes_match.match_length] += 1
        txt = bytes_match.__rich__()

        if bytes_match.match_length < YaralyzerConfig.MIN_DECODE_LENGTH:
            txt = Text('Too little to actually attempt decode at ', style='grey') + txt
        else:
            txt.append(" too long to decode ")
            txt.append(f"(--max-decode-length is {YaralyzerConfig.MAX_DECODE_LENGTH} bytes)", style='grey')

        log.debug(Text('Queueing suppression notice: ') + txt)
        self.suppression_notice_queue.append(txt)

    def _print_suppression_notices(self) -> None:
        """Print notices in queue in a single panel; empty queue"""
        if len(self.suppression_notice_queue) == 0:
            return

        suppression_notices_txt = Text("\n").join([notice for notice in self.suppression_notice_queue])
        panel = Panel(suppression_notices_txt, style='bytes', expand=False)
        console.print(panel)
        self.suppression_notice_queue = []

    def _eexec_idx(self) -> int:
        """Returns the location of CURRENTFILES_EEXEC within the binary stream dataor 0"""
        return self.bytes.find(CURRENTFILE_EEXEC) if CURRENTFILE_EEXEC in self.bytes else 0


def new_decoding_stats_table(title) -> Table:
    """Build an empty table for displaying decoding stats"""
    title = prefix_with_plain_text_obj(title, style='blue underline')
    title.append(": Decoding Attempts Summary Statistics", style='bright_white bold')

    table = Table(
        title=title,
        min_width=half_width(),
        show_lines=True,
        padding=(0, 1),
        style='color(18)',
        border_style='color(111) dim',
        header_style='color(235) on color(249) reverse',
        title_style='color(249) bold')

    def add_column(header, **kwargs):
        table.add_column(pad_header(header.upper()), **kwargs)

    add_column('Byte Pattern', vertical='middle', style='color(25) bold reverse', justify='right')
    add_column('Aggregate Metrics', overflow='fold', justify=CENTER)
    add_column('Per Encoding Metrics', justify=CENTER)
    return table



"""
There's two possible log sinks other than STDOUT:

  1. 'log' - the application log (standard log, what goes to STDOUT with -D option)
  2. 'invocation_log' - tracks the exact command pdfalyzer was invoked with, similar to a history file

The regular log file at APPLICATION_LOG_PATH is where the quite verbose application logs
will be written if things ever need to get that formal. For now those logs are only accessible
on STDOUT with the -D flag but the infrastructure for persistent logging exists if someone
needs/wants that sort of thing.

Logs are not normally ephemeral/not written  to files but can be configured to do so by setting
the PDFALYZER_LOG_DIR env var. See .pdfalyzer.example for documentation about the side effects of setting
PDFALYZER_LOG_DIR to a value.

https://docs.python.org/3/library/logging.html#logging.basicConfig
https://realpython.com/python-logging/

Python log levels for reference:
    CRITICAL 50
    ERROR 40
    WARNING 30
    INFO 20
    DEBUG 10
    NOTSET 0
"""

import logging
from os import environ, path
from rich.logging import RichHandler

from pdfalyzer.config import PdfalyzerConfig


def configure_logger(log_label: str) -> logging.Logger:
    """Set up a file or stream logger depending on the configuration"""
    log_name = f"pdfalyzer.{log_label}"
    logger = logging.getLogger(log_name)

    if PdfalyzerConfig.LOG_DIR:
        if not path.isdir(PdfalyzerConfig.LOG_DIR) or not path.isabs(PdfalyzerConfig.LOG_DIR):
            raise RuntimeError(f"PDFALYZER_LOG_DIR '{PdfalyzerConfig.LOG_DIR}' doesn't exist or is not absolute")

        log_file_path = path.join(PdfalyzerConfig.LOG_DIR, f"{log_name}.log")
        log_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        log_file_handler = logging.FileHandler(log_file_path)
        log_file_handler.setFormatter(log_formatter)
        logger.addHandler(log_file_handler)
        # rich_stream_handler is for printing warnings
        rich_stream_handler = RichHandler(rich_tracebacks=True)
        rich_stream_handler.setLevel('WARN')
        logger.addHandler(rich_stream_handler)
        logger.info('File logging triggered by setting of PDFALYZER_LOG_DIR')
    else:
        logger.addHandler(RichHandler(rich_tracebacks=True))

    logger.setLevel(PdfalyzerConfig.LOG_LEVEL)
    return logger


# See comment at top. 'log' is the standard application log, 'invocation_log' is a history of pdfalyzer runs
log = configure_logger('run')
invocation_log = configure_logger('invocation')

# If we're logging to files make sure invocation_log has the right level
if PdfalyzerConfig.LOG_DIR:
    invocation_log.setLevel('INFO')


def log_and_print(msg: str, log_level='INFO'):
    """Both print and log (at INFO level) a string"""
    log.log(logging.getLevelName(log_level), msg)
    print(msg)


def log_current_config():
    """Write current state of PdfalyzerConfig object to the logs"""
    msg = f"{PdfalyzerConfig.__name__} current attributes:\n"
    config_dict = {k: v for k, v in vars(PdfalyzerConfig).items() if not k.startswith('__')}

    for k in sorted(config_dict.keys()):
        msg += f"   {k: >35}  {config_dict[k]}\n"

    log.info(msg)


# Suppress annoying chardet library logs
for submodule in ['universaldetector', 'charsetprober', 'codingstatemachine']:
    logging.getLogger(f"chardet.{submodule}").setLevel(logging.WARNING)



"""
Some methods to help with the direct manipulation/processing of PyPDF2's PdfObjects
"""
from collections import namedtuple
from typing import List

from PyPDF2.generic import IndirectObject, PdfObject
from rich.markup import escape
from yaralyzer.util.logging import log

from pdfalyzer.output.layout import get_label_style
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

# In the case of easy key/value pairs the reference_key and the reference_address are the same but
# for more complicated references the reference_address will be the reference_key plus sub references.
#
#   e.g. a reference_key for a /Font labeled /F1 could be '/Resources' but the reference_address
# might be '/Resources[/Font][/F1] if the /Font is a directly embedded reference instead of a remote one.
PdfObjectRef = namedtuple('PdfObjectRef', ['reference_key', 'reference_address', 'pdf_obj'])

# For printing SymlinkNodes
SymlinkRepresentation = namedtuple('SymlinkRepresentation', ['text', 'style'])


def get_references(obj: PdfObject, ref_key=None, ref_address=None) -> List[PdfObjectRef]:
    """Return list of PdfObjectRefs"""
    if isinstance(obj, IndirectObject):
        if ref_key is None or ref_address is None:
            raise PdfWalkError(f"{obj} is a reference but key or address not provided")

        return [PdfObjectRef(ref_key, ref_address, obj)]
    elif isinstance(obj, list) and isinstance(obj, dict):
        raise PdfWalkError(f"PdfObject {obj} is both a dict and a list")

    return_list = []

    if isinstance(obj, list):
        for i, element in enumerate(obj):
            if not isinstance(element, (IndirectObject, list, dict)):
                continue

            _ref_address = f"[{i}]" if ref_address is None else f"{ref_address}[{i}]"
            return_list += get_references(element, ref_key or i, _ref_address)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            _ref_address = k if ref_address is None else f"{ref_address}[{k}]"
            return_list += get_references(v, ref_key or k, _ref_address)
    else:
        log.debug(f"Adding no references for PdfObject reference '{ref_key}' -> '{obj}'")

    return return_list


def get_symlink_representation(from_node, to_node) -> SymlinkRepresentation:
    """Returns a tuple (symlink_text, style) that can be used for pretty printing, tree creation, etc"""
    reference_key = str(to_node.get_reference_key_for_relationship(from_node))
    pdf_instruction = reference_key.split('[')[0]  # In case we ended up with a [0] or similar

    if pdf_instruction in DANGEROUS_PDF_KEYS:
        symlink_style = 'red_alert'
    else:
        symlink_style = get_label_style(to_node.label) + ' dim'

    symlink_str = f"{escape(reference_key)} [bright_white]=>[/bright_white]"
    symlink_str += f" {escape(str(to_node.target))} [grey](Non Child Reference)[/grey]"
    return SymlinkRepresentation(symlink_str, symlink_style)


def pdf_object_id(pdf_object):
    """Return the ID of an IndirectObject and None for everything else"""
    return pdf_object.idnum if isinstance(pdf_object, IndirectObject) else None


def does_list_have_any_references(_list) -> bool:
    """Return true if any element of _list is an IndirectObject"""
    return any(isinstance(item, IndirectObject) for item in _list)



"""
Rich colors: https://rich.readthedocs.io/en/stable/appendix/colors.html
TODO: interesting colors # row_styles[0] = 'reverse bold on color(144)' <-
"""
import re
import time
from numbers import Number
from os import path
from typing import List, Union

from PyPDF2.generic import ByteStringObject, IndirectObject
from rich import box
from rich.columns import Columns
from rich.padding import Padding
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from yaralyzer.output.rich_console import GREY_ADDRESS, YARALYZER_THEME_DICT, console
from yaralyzer.helpers.rich_text_helper import prefix_with_plain_text_obj
from yaralyzer.util.logging import log, log_and_print

from pdfalyzer.helpers.dict_helper import merge

# Colors
DANGER_HEADER = 'color(88) on white'  # Red

# PDF object styles
PDF_ARRAY = 'color(120)'
PDF_NON_TREE_REF = 'color(243)'

PDFALYZER_THEME_DICT = YARALYZER_THEME_DICT.copy()
PDFALYZER_THEME_DICT.update({
    'address': GREY_ADDRESS,
    'BOM': 'bright_green',
    # PDF objects
    'pdf.array': PDF_ARRAY,
    'pdf.non_tree_ref': PDF_NON_TREE_REF,
    # fonts
    'font.property': 'color(135)',
    'font.title': 'reverse dark_blue on color(253)',
    # charmap
    'charmap.title': 'color(18) reverse on white dim',
    'charmap.prepared_title': 'color(23) reverse on white dim',
    'charmap.prepared': 'color(106) dim',
    'charmap.byte': 'color(58)',
    'charmap.char': 'color(120) bold',
    # design elements
    'subtable': 'color(8) on color(232)',
    'header.minor': 'color(249) bold',
    'header.danger': DANGER_HEADER,
    'header.danger_reverse': f'{DANGER_HEADER} reverse',
    # neutral log events
    'event.attn': 'bold bright_cyan',
    'event.lowpriority': 'bright_black',
    # good log events
    'event.good': 'green4',
    'event.better': 'turquoise4',
    'event.reallygood': 'dark_cyan',
    'event.reallygreat': 'spring_green1',
    'event.great': 'sea_green2',
    'event.evenbetter': 'chartreuse1',
    'event.best': 'green1',
    'event.siren': 'blink bright_white on red3',
    # warn log events
    'warn': 'bright_yellow',
    'warn.mild': 'yellow2',
    'warn.milder': 'dark_orange3',
    'warn.harsh': 'reverse bright_yellow',
    # error log events
    'fail': 'bold reverse red',
    'milderror': 'red',
    'red_alert': 'blink bold red reverse on white',
})

console.push_theme(Theme(PDFALYZER_THEME_DICT))


TYPE_STYLES = {
    Number: 'bright_cyan bold',
    dict: 'color(64)',
    list: 'color(143)',
    str: 'bright_white bold',
    IndirectObject: 'color(157)',
    ByteStringObject: 'bytes',
}

# Color meter realted constants. Make even sized buckets color coded from blue (cold) to green (go)
METER_COLORS = list(reversed([82, 85, 71, 60, 67, 30, 24, 16]))
METER_INTERVAL = (100 / float(len(METER_COLORS))) + 0.1

# Color meter extra style thresholds (these are assuming a scale of 0-100)
UNDERLINE_CONFIDENCE_THRESHOLD = 90
BOLD_CONFIDENCE_THRESHOLD = 60
DIM_COUNTRY_THRESHOLD = 25

# Table stuff
DEFAULT_SUBTABLE_COL_STYLES = ['white', 'bright_white']
HEADER_PADDING = (1, 1)

# For the table shown by running pdfalyzer_show_color_theme
MAX_THEME_COL_SIZE = 35

# Text object defaults mostly for table entries
NOT_FOUND_MSG = Text('(not found)', style='grey.dark_italic')


def get_type_style(klass) -> str:
    """Style for various types of data (e.g. DictionaryObject)"""
    return next((TYPE_STYLES[t] for t in TYPE_STYLES.keys() if issubclass(klass, t)), None)


def get_type_string_style(klass) -> str:
    """Dim version of get_type_style() for non primitives, white for primitives"""
    if issubclass(klass, (str, Number)):
        return 'white'
    else:
        return f"{get_type_style(klass)} dim"


def generate_subtable(cols, header_style='subtable') -> Table:
    """Suited for lpacement in larger tables"""
    table = Table(
        box=box.SIMPLE,
        show_edge=False,
        collapse_padding=True,
        header_style=header_style,
        show_lines=False,
        border_style='grey.dark',
        expand=True)

    for i, col in enumerate(cols):
        if i + 1 < len(cols):
            table.add_column(col, style=DEFAULT_SUBTABLE_COL_STYLES[0], justify='left')
        else:
            table.add_column(col, style='off_white', justify='right')

    return table


def pad_header(header: str) -> Padding:
    """Would pad anything, not just headers"""
    return Padding(header, HEADER_PADDING)


def pdfalyzer_show_color_theme() -> None:
    """Utility method to show pdfalyzer's color theme. Invocable with 'pdfalyzer_show_colors'."""
    console.print(Panel('The Pdfalyzer Color Theme', style='reverse'))

    colors = [
        prefix_with_plain_text_obj(name[:MAX_THEME_COL_SIZE], style=str(style)).append(' ')
        for name, style in PDFALYZER_THEME_DICT.items()
        if name not in ['reset', 'repr_url']
    ]

    console.print(Columns(colors, column_first=True, padding=(0,3)))



"""
Class to help with the pre-configured YARA rules in /yara.
"""
from importlib.resources import as_file, files
from typing import Optional, Union

from yaralyzer.yaralyzer import Yaralyzer

YARA_RULES_DIR = files('pdfalyzer').joinpath('yara_rules')

YARA_RULES_FILES = [
    'lprat.static_file_analysis.yara',
    'PDF.yara',
    'PDF_binary_stream.yara',
]

def get_file_yaralyzer(file_path_to_scan: str) -> Yaralyzer:
    """Get a yaralyzer for a file path"""
    return _build_yaralyzer(file_path_to_scan)


def get_bytes_yaralyzer(scannable: bytes, label: str) -> Yaralyzer:
    return _build_yaralyzer(scannable, label)


def _build_yaralyzer(scannable: Union[bytes, str], label: Optional[str] = None) -> Yaralyzer:
    # TODO: ugh this sucks (handling to extract .yara files from a python pkg zip)
    with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[0])) as yara0:
        with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[1])) as yara1:
            with as_file(YARA_RULES_DIR.joinpath(YARA_RULES_FILES[2])) as yara2:
                rules_paths = [str(y) for y in [yara0, yara1, yara2]]
                return Yaralyzer.for_rules_files(rules_paths, scannable, label)



"""
Various text formatting/styling/manipulating methods.
"""
import re
from pprint import PrettyPrinter
from typing import Pattern

from PyPDF2.generic import PdfObject
from yaralyzer.output.rich_console import console_width


# Style
INDENT_DEPTH = 4
PRETTY_PRINT_WIDTH = 60


# Pretty Printer
pp = PrettyPrinter(
    indent=INDENT_DEPTH,
    width=PRETTY_PRINT_WIDTH,
    sort_dicts=True)


def pypdf_class_name(obj: PdfObject) -> str:
    """Shortened name of type(obj), e.g. PyPDF2.generic._data_structures.ArrayObject becomes Array"""
    class_pkgs = type(obj).__name__.split('.')
    class_pkgs.reverse()
    return class_pkgs[0].removesuffix('Object')


def generate_hyphen_line(width=None, title=None):
    """e.g. '-----------------BEGIN-----------------'"""
    width = width or console_width()

    if title is None:
        return '-' * width

    side_hyphens = int((width - len(title)) / 2) * '-'
    line = side_hyphens + title + side_hyphens
    return line if len(line) == width else line + '-'


def list_to_string(_list: list, sep=', ') -> str:
    """Join elements of _list with sep"""
    return sep.join([str(item) for item in _list])


def count_pattern_matches_in_text(pattern: str, text: str) -> int:
    return count_regex_matches_in_text(re.compile(pattern), text)


def count_regex_matches_in_text(regex: Pattern, text: str) -> int:
    """For use when you precompile the regex"""
    return sum(1 for _ in regex.finditer(text))



"""
String constants specified in the Adobe specs for PDFs, fonts, etc.
"""
import re

from PyPDF2.constants import (CatalogDictionary, ImageAttributes, PageAttributes,
     PagesAttributes, Ressources as Resources)

# Fake PDF instructions used to create more explanatory tables/trees/addresses/etc.
ARRAY_ELEMENT = '/ArrayElement'
TRAILER = '/Trailer'
UNLABELED = '/UnlabeledArrayElement'

# Actual PDF instructions
AA              = CatalogDictionary.AA  # Automatic Action
ACRO_FORM       = CatalogDictionary.ACRO_FORM  # Can trigger Javascript on open
ANNOTS          = '/Annots'
COLOR_SPACE     = Resources.COLOR_SPACE
D               = '/D'  # Destination, usually of a link or action
CONTENTS        = '/Contents'
DEST            = '/Dest'  # Similar to /D?
EXT_G_STATE     = Resources.EXT_G_STATE
FIELDS          = '/Fields'
FIRST           = '/First'
FONT            = Resources.FONT
FONT_FILE       = '/FontFile'
FONT_FILE2      = FONT_FILE + '2'
FONT_FILE3      = FONT_FILE + '3'
FONT_DESCRIPTOR = '/FontDescriptor'
GROUP           = '/Group'
JAVASCRIPT      = '/JavaScript'
JS              = '/JS'
K               = '/K'  # Equivalent of /Kids for /StructElem
KIDS            = PagesAttributes.KIDS
LAST            = '/Last'
NAMES           = '/Names'
NEXT            = '/Next'
NUMS            = '/Nums'
OBJECT_STREAM   = '/ObjStm'
OBJ             = '/Obj'
# TODO: /Pg refs could be the parents of /OBJR?
OBJR            = '/OBJR'  # Object reference to "an entire PDF object"
OPEN_ACTION     = CatalogDictionary.OPEN_ACTION
P               = '/P'  # Equivalent of /Parent for /StructElem
PAGE            = '/Page'
PAGES           = '/Pages'
PARENT          = PagesAttributes.PARENT
PG              = '/Pg'  # Page ref for OBJR
PREV            = '/Prev'
RESOURCES       = PageAttributes.RESOURCES
S               = '/S'  # Equivalent of /Subtype for /StructElem
SIZE            = '/Size'
STRUCT_ELEM     = '/StructElem'
SUBTYPE         = ImageAttributes.SUBTYPE
TO_UNICODE      = '/ToUnicode'
TYPE            = PageAttributes.TYPE
TYPE1_FONT      = '/Type1'
W               = '/W'  # Equivalen of /Widths in some situations
WIDGET          = '/Widget'
WIDTHS          = '/Widths'
XOBJECT         = Resources.XOBJECT
XREF            = '/XRef'
XREF_STREAM     = '/XRefStm'

# There can be up to 3 /Length1, Length2, etc. keys depending on the type of font.
# They indicate points in the binary stream where different sections of the font definition
# can be found.
FONT_LENGTHS = [f'/Length{i + 1}' for i in range(3)]

# Instructions to flag when scanning stream data for malicious content.
DANGEROUS_PDF_KEYS = [
    # AA,  # AA is too generic; can't afford to remove the frontslash
    ACRO_FORM,
    JAVASCRIPT,
    JS,
    OPEN_ACTION
]

# Adobe font instruction that begins the binary (usually encrypted) section of the font definition
CURRENTFILE_EEXEC = b'currentfile eexec'

# A node with this label is really just a non-tree link between nodes
PURE_REFERENCE_NODE_LABELS = [
    D,
    DEST,
    NUMS
]

# Some references are never part of a parent/child relationship in the tree
NON_TREE_REFERENCES = [
    OPEN_ACTION,
#    D,
    LAST,
    NEXT,
    PREV,
]

# Some PdfObjects can't be properly placed in the tree until the entire tree is parsed
INDETERMINATE_REFERENCES = [
    ANNOTS,  # At least when it appears in a page
    COLOR_SPACE,
    D,
    DEST,
    EXT_G_STATE,
    FIELDS,   # At least for  /AcroForm
    FIRST,
    FONT,
    NAMES,
    OPEN_ACTION,
    P,   # At least for widgets...
    RESOURCES,
    XOBJECT,
    UNLABELED, # TODO: this might be wrong? maybe this is where the /Resources actually live?
]

EXTERNAL_GRAPHICS_STATE_REGEX = re.compile('/Resources\\[/ExtGState\\]\\[/GS\\d+\\]')



"""
Deprecated pre-tree, more rawformat reader.
"""
from io import StringIO

from PyPDF2.generic import ArrayObject, DictionaryObject, IndirectObject
from rich.console import Console
from rich.markup import escape

from pdfalyzer.helpers.string_helper import pp, pypdf_class_name, INDENT_DEPTH


INDENT_SPACES = ' ' * INDENT_DEPTH
INDENT_JOIN = "\n" + INDENT_SPACES
VALUE_WIDTH = 30

TRUNCATABLE_TYPES = (ArrayObject, DictionaryObject, list, dict)
TRUNCATE_MULTILINE = 25


# Truncates large pretty print output
def pretty_print_list_or_dict(obj):
    if isinstance(obj, list) and len(obj) > TRUNCATE_MULTILINE:
        truncate_msg = f'(truncated {len(obj) - TRUNCATE_MULTILINE} of {len(obj)} rows)'
        obj = obj[:TRUNCATE_MULTILINE] + [truncate_msg]

    pretty_str = pp.pformat(obj)

    if pretty_str.count("\n") > TRUNCATE_MULTILINE:
        pretty_str = "\n".join(pretty_str.split("\n")[:TRUNCATE_MULTILINE])
        pretty_str += f"{INDENT_JOIN}(...truncated...)"

    return pretty_str


def print_all_props(pdf_obj, console, verbose=False, indent=''):
    if 'keys' not in dir(pdf_obj):
        console.print('* ' + pp.pformat(pdf_obj))
        return

    # Save large object output for last
    large_print_jobs = {}

    for k, v in pdf_obj.items():
        if (verbose and isinstance(v, IndirectObject)):
            v = v.get_object()

        k = pp.pformat(k) if isinstance(k, tuple) else k
        pretty_str = pp.pformat(v)

        if (len(pretty_str) > VALUE_WIDTH and (isinstance(v, TRUNCATABLE_TYPES))):
            large_print_jobs[k] = v
            continue

        pretty_str = '{0: <20}{1: <{vwidth}}'.format(k, pretty_str, vwidth=VALUE_WIDTH)
        console.print(indent + ' {0: <66}{1: >20}'.format(pretty_str, pypdf_class_name(v)))

    for k, v in large_print_jobs.items():
        indent_join = INDENT_JOIN + indent
        pretty_str = indent_join.join(pretty_print_list_or_dict(v).split("\n"))
        console.print(indent + f" {k}  ({pypdf_class_name(v)})    {indent_join}{pretty_str}")


# Prints with a header of your choosing
def print_with_header(obj, header, depth=0, print_props=True, print_header=True):
    console = Console(file=StringIO())
    box_horiz = '-' * (len(header) + 4)
    box_elements = [box_horiz, f'| {escape(header)} |', box_horiz]
    indent = INDENT_SPACES * depth
    indent_join = "\n" + indent

    if print_header:
        console.print(f'{indent}' + indent_join.join(box_elements))

        if print_props:
            print_all_props(obj, console, indent=indent)

    console.print('')
    return console.file.getvalue()



import importlib.resources
import logging
from os import environ, pardir, path

from yaralyzer.config import YaralyzerConfig, is_env_var_set_and_not_false, is_invoked_by_pytest

PDFALYZE = 'pdfalyze'
PROJECT_ROOT = path.join(str(importlib.resources.files('pdfalyzer')), pardir)
PYTEST_FLAG = 'INVOKED_BY_PYTEST'

# Configuring PDFALYZER_LOG_DIR has side effects; see .pdfalyzer.example in repo for specifics.
LOG_LEVEL_ENV_VAR = 'PDFALYZER_LOG_LEVEL'
LOG_DIR_ENV_VAR = 'PDFALYZER_LOG_DIR'

# 3rd part pdf-parser.py
PDF_PARSER_EXECUTABLE_ENV_VAR = 'PDFALYZER_PDF_PARSER_PY_PATH'


YaralyzerConfig.LOG_DIR = environ.get(LOG_DIR_ENV_VAR)
YaralyzerConfig.LOG_LEVEL = logging.getLevelName(environ.get(LOG_LEVEL_ENV_VAR, 'WARN'))


class PdfalyzerConfig:
    JAVSCRIPT_KEYWORD_ALERT_THRESHOLD = 2
    QUOTE_TYPE = None
    DEFAULT_PDF_PARSER_EXECUTABLE = path.join(PROJECT_ROOT, 'tools', 'pdf-parser.py')

    # Path to Didier Stevens's pdf-parser.py
    if is_env_var_set_and_not_false(PDF_PARSER_EXECUTABLE_ENV_VAR):
        PDF_PARSER_EXECUTABLE = path.join(environ[PDF_PARSER_EXECUTABLE_ENV_VAR], 'pdf-parser.py')
    elif is_invoked_by_pytest():
        PDF_PARSER_EXECUTABLE = DEFAULT_PDF_PARSER_EXECUTABLE
    else:
        if path.exists(DEFAULT_PDF_PARSER_EXECUTABLE):
            PDF_PARSER_EXECUTABLE = DEFAULT_PDF_PARSER_EXECUTABLE
        else:
            PDF_PARSER_EXECUTABLE = None



"""
PDF object decorator - wraps actual PDF objects to make them anytree nodes.
Also adds decorators/generators for Rich text representation.

Child/parent relationships should be set using the add_child()/set_parent()
methods and not set directly. (TODO: this could be done better with anytree
hooks)
"""
from collections import namedtuple
from numbers import Number
from typing import List, Optional, Union

from anytree import NodeMixin, SymlinkNode
from PyPDF2.generic import (DictionaryObject, EncodedStreamObject, IndirectObject, NumberObject, PdfObject,
     StreamObject)
from PyPDF2.errors import PdfReadError
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree
from yaralyzer.helpers.bytes_helper import clean_byte_string, hex_text
from yaralyzer.helpers.rich_text_helper import size_text
from yaralyzer.output.rich_console import BYTES_NO_DIM, YARALYZER_THEME, console
from yaralyzer.util.logging import log

from yaralyzer.encoding_detection.character_encodings import NEWLINE_BYTE
from pdfalyzer.helpers.pdf_object_helper import PdfObjectRef, get_references, get_symlink_representation
from pdfalyzer.helpers.rich_text_helper import (PDF_ARRAY, TYPE_STYLES,
     get_type_style, get_type_string_style)
from pdfalyzer.helpers.string_helper import pypdf_class_name
from pdfalyzer.output.layout import get_label_style
from pdfalyzer.util.adobe_strings import *
from pdfalyzer.util.exceptions import PdfWalkError

Relationship = namedtuple('Relationship', ['from_node', 'reference_key'])

STREAM_PREVIEW_LENGTH_IN_TABLE = 500
DEFAULT_MAX_ADDRESS_LENGTH = 90
STREAM = 'Stream'
HEX = 'Hex'
PREVIEW_STYLES = {HEX: BYTES_NO_DIM, STREAM: 'bytes'}


class PdfTreeNode(NodeMixin):
    def __init__(self, obj: PdfObject, address: str, idnum: int):
        """
        reference_key: PDF instruction string used to reference obj
        idnum: ID used in the reference
        """
        self.obj = obj
        self.idnum = idnum
        self.other_relationships = []
        self.all_references_processed = False

        if isinstance(obj, StreamObject):
            try:
                self.stream_data = self.obj.get_data()
            except PdfReadError as e:
                msg = f"Failed to decode stream: {e}"
                self.stream_data = msg
                log.error(msg)
                console.print_exception()

            self.stream_length = len(self.stream_data)
        else:
            self.stream_data = None
            self.stream_length = 0

        if isinstance(obj, DictionaryObject):
            self.type = obj.get(TYPE) or address
            self.label = obj.get(TYPE) or address
            self.sub_type = obj.get(SUBTYPE) or obj.get(S)

            if isinstance(self.type, str):
                self.type = self.type.split('[')[0]
        else:
            # TODO: hacky
            self.type = address.split('[')[0] if isinstance(address, str) else address
            self.label = address
            self.sub_type = None

        # TODO: this is hacky and possibly incorrect
        if isinstance(address, int):
            self.known_to_parent_as = f"[{address}]"
        else:
            self.known_to_parent_as = address

        if isinstance(self.label, int):
            self.label = f"{UNLABELED}[{self.label}]"

    @classmethod
    def from_reference(cls, ref: IndirectObject, address: str) -> 'PdfTreeNode':
        """Builds a PdfTreeDecorator from an IndirectObject"""
        return cls(ref.get_object(), address, ref.idnum)

    def set_parent(self, parent: 'PdfTreeNode') -> None:
        """Set the parent of this node"""
        if self.parent and self.parent != parent:
            raise PdfWalkError(f"Cannot set {parent} as parent of {self}, parent is already {self.parent}")

        self.parent = parent
        self.remove_relationship(parent)
        # Adjust incorrect known_to_parent_as that can arise when adding non tree references to the walker
        self.known_to_parent_as = self._find_address_of_this_node(parent) or self.known_to_parent_as
        log.info(f"  Added {parent} as parent of {self}")

    def add_child(self, child: 'PdfTreeNode') -> None:
        """Add a child to this node"""
        existing_child = next((c for c in self.children if c.idnum == child.idnum), None)

        if existing_child is not None:
            if existing_child == child:
                log.debug(f"{child} is already child of {self}")
                return
            else:
                raise PdfWalkError(f"{self} already has child w/this ID: {child}")

        self.children += (child,)
        child.remove_relationship(self)
        # Adjust incorrect known_to_parent_as that can arise when adding non tree references to the walker
        child.known_to_parent_as = child._find_address_of_this_node(self) or child.known_to_parent_as
        log.info(f"  Added {child} as child of {self}")

    def add_relationship(self, from_node: 'PdfTreeNode', reference_key: str) -> None:
        """Handles relationships not covered by the tree structure"""
        relationship = Relationship(from_node, reference_key)

        if relationship in self.other_relationships:
            return

        log.info(f'Adding other relationship: {from_node} ref {reference_key} points to {self}')
        self.other_relationships.append(relationship)

    def remove_relationship(self, from_node: 'PdfTreeNode') -> None:
        """Remove all other_relationships from from_node to this node"""
        relationships_to_remove = [r for r in self.other_relationships if r.from_node == from_node]

        if len(relationships_to_remove) == 0:
            return
        elif len(relationships_to_remove) > 1 and not all(r.reference_key in [FIRST, LAST] for r in relationships_to_remove):
            log.warning(f"> 1 relationships to remove from {from_node} to {self}: {relationships_to_remove}")

        for relationship in relationships_to_remove:
            log.debug(f"Removing relationship {relationship} from {self}")
            self.other_relationships.remove(relationship)

    def other_relationship_count(self) -> int:
        return len(self.other_relationships)

    def get_reference_key_for_relationship(self, from_node: 'PdfTreeNode'):
        """Get the label that links from_node to this one outside of the tree structure"""
        relationship = next((r for r in self.other_relationships if r.from_node == from_node), None)

        if relationship is None:
            raise PdfWalkError(f"Can't find other relationship from {from_node} to {self}")

        return relationship.reference_key

    # TODO: this doesn't include /Parent references
    def referenced_by_keys(self) -> List[str]:
        """All the PDF instruction strings that referred to this object"""
        return [r.reference_key for r in self.other_relationships] + [self.known_to_parent_as]

    # Old clause:  elif key in NON_TREE_REFERENCES or node.label.startswith(NUMS) or node.label in PURE_REFERENCE_NODE_LABELS:
    def is_pure_reference(self, reference_key) -> bool:
        """Returns True if the reference is probably not in the tree"""
        # if self.idnum == 505:
        #     import pdb;pdb.set_trace()
        if reference_key in (NON_TREE_REFERENCES + PURE_REFERENCE_NODE_LABELS):
            return True
#        if isinstance(reference_key, int):
#            return False

        # TODO: Checking startswith(NUMS) etc. is a hack that probably will not cover all cases with /StructElem
        return any(self.label.startswith(key) for key in PURE_REFERENCE_NODE_LABELS)

    def is_parent_reference(self, reference_key) -> bool:
        """Returns True for explicit parent refs"""
        if reference_key == PARENT:
            return True
        elif self.type == STRUCT_ELEM and reference_key == P:
            return True
        else:
            return False

    def is_child_reference(self, reference_key) -> bool:
        """Returns True for explicit child refs"""
        if reference_key == KIDS:
            return True
        elif self.type == STRUCT_ELEM and reference_key == K:
            return True
        elif self.type == OBJR and reference_key == OBJ:
            # TODO: there can be multiple OBJR refs to the same object... which wouldn't work w/this code
            return True
        else:
            return False

    # old check: elif key in INDETERMINATE_REFERENCES and key == address:
    # TODO: why did we check equality?
    def is_indeterminate_reference(self, reference_key):
        """Returns true if we need to wait for all objects to be parsed before placement"""
        if reference_key in INDETERMINATE_REFERENCES:
            return True
        else:
            return False

    def references(self) -> List[PdfObjectRef]:
        """Returns all nodes referenced from this node (see PdfObjectRef definition)"""
        return get_references(self.obj)

    def contains_stream(self) -> bool:
        """Returns True for ContentStream, DecodedStream, and EncodedStream objects"""
        return isinstance(self.obj, StreamObject)

    def _find_address_of_this_node(self, other_node: 'PdfTreeNode') -> Optional[str]:
        """Find the address used in other_node to refer to this node"""
        refs_to_this_node = [ref for ref in other_node.references() if ref.pdf_obj.idnum == self.idnum]

        if len(refs_to_this_node) == 1:
            return refs_to_this_node[0].reference_address
        elif len(refs_to_this_node) == 0:
            # /XRef streams are basically trailer nodes without any direct reference
            if self.parent.label == TRAILER and self.type == XREF and XREF_STREAM in self.parent.obj:
                return XREF_STREAM
            elif self.label not in [FIRST, LAST, NEXT, PREV]:
                log.warning(f"Could not find expected reference from {other_node} to {self}")
                return None
        else:
            reference_address = refs_to_this_node[0].reference_address

            if not all(ref.reference_address in [FIRST, LAST] for ref in refs_to_this_node):
                log.warning(f"Multiple refs from {other_node} to {self}: {refs_to_this_node}. Using {reference_address} as address")

            return reference_address

    ######################################
    # BELOW HERE IS JUST TEXT FORMATTING #
    ######################################

    def print_other_relationships(self) -> None:
        """Print this node's non tree relationships (the ones represented by SymlinkNodes in the tree)"""
        console.print(f"Other relationships of {escape(str(self))}")

        for r in self.other_relationships:
            console.print(f"  Referenced as {escape(str(r.reference_key))} by {escape(str(r.from_node))}")

    def tree_address(self, max_length: Union[int, None]  = None) -> str:
        """Creates a string like '/Catalog/Pages/Resources/Font' truncated to max_length (if given)"""
        if self.label == TRAILER:
            return '/'

        address = ''.join([str(node.known_to_parent_as) for node in self.path if node.label != TRAILER])
        address_length = len(address)

        if max_length is None or max_length > address_length:
            return address

        return '...' + address[-max_length:][3:]

    def colored_address(self, max_length: Union[int, None] = None) -> Text:
        """Rich text version of tree_address()"""
        text = Text('@', style='bright_white')
        text.append(self.tree_address(max_length), style='address')
        return text

    def colored_node_label(self) -> Text:
        return Text(self.label[1:], style=f'{get_label_style(self.label)} underline bold')

    def colored_node_type(self) -> Text:
        return Text(pypdf_class_name(self.obj), style=get_node_type_style(self.obj))

    def generate_rich_table(self) -> Table:
        """
        Generate a Rich table representation of this node's PDF object and its properties.
        Table cols are [title, address, class name] (not exactly headers but sort of).
        """
        title = f"{self.idnum}.{escape(self.label)}"
        table = Table(title, escape(self.tree_address()), pypdf_class_name(self.obj))
        table.columns[0].header_style = f'reverse {get_label_style(self.label)}'
        table.columns[1].header_style = 'dim'
        table.columns[1].overflow = 'fold'
        table.columns[2].header_style = get_node_type_style(self.obj)

        if self.label != self.known_to_parent_as:
            table.add_row(Text('ParentRefKey', style='grey'), Text(str(self.known_to_parent_as), style='grey'), '')

        if isinstance(self.obj, dict):
            for k, v in self.obj.items():
                row = to_table_row(k, v)

                # Make dangerous stuff look dangerous
                if (k in DANGEROUS_PDF_KEYS) or (self.label == FONT and k == SUBTYPE and v == TYPE1_FONT):
                    table.add_row(*[col.plain for col in row], style='fail')
                else:
                    table.add_row(*row)
        elif isinstance(self.obj, list):
            for i, item in enumerate(self.obj):
                table.add_row(*to_table_row(i, item))
        elif not isinstance(self.obj, EncodedStreamObject):
            # Then it's a single element node like a URI, TextString, etc.
            table.add_row(*to_table_row('', self.obj, is_single_row_table=True))

        for row in self._get_stream_preview_rows():
            row.append(Text(''))
            table.add_row(*row)

        return table

    def generate_rich_tree(self, tree=None, depth=0):
        """Recursively generates a rich.tree.Tree object from this node"""
        tree = tree or Tree(self.generate_rich_table())

        for child in self.children:
            if isinstance(child, SymlinkNode):
                symlink_rep = get_symlink_representation(self, child)
                tree.add(Panel(symlink_rep.text, style=symlink_rep.style, expand=False))
                continue

            child_branch = tree.add(child.generate_rich_table())
            child.generate_rich_tree(child_branch)

        return tree

    def _get_stream_preview_rows(self) -> List[List[Text]]:
        """Get rows that preview the stream data"""
        return_rows: List[List[Text]] = []

        if not isinstance(self.obj, StreamObject):
            return return_rows
        if self.stream_data is None or len(self.stream_data) == 0:
            log.warning(self.__rich__().append(' is a stream object but had no stream data'))
            return return_rows

        stream_preview = self.stream_data[:STREAM_PREVIEW_LENGTH_IN_TABLE]
        stream_preview_length = len(stream_preview)

        if isinstance(self.stream_data, bytes):
            stream_preview_hex = hex_text(stream_preview).plain
        else:
            stream_preview_hex = f"N/A (Stream data is type '{type(self.stream_data).__name__}', not bytes)"

        if isinstance(stream_preview, bytes):
            stream_preview_lines = stream_preview.split(NEWLINE_BYTE)
            stream_preview_string = "\n".join([clean_byte_string(line) for line in stream_preview_lines])
        else:
            stream_preview_string = stream_preview

        def add_preview_row(hex_or_stream: str, stream_string: str):
            if stream_preview_length < STREAM_PREVIEW_LENGTH_IN_TABLE:
                row_label = "Data" if hex_or_stream != HEX else ' View'
            else:
                row_label = "Preview" if hex_or_stream != HEX else ' Preview'
                stream_string += '...'

            style = PREVIEW_STYLES[hex_or_stream]
            row_label = f"{hex_or_stream}{row_label}\n  ({stream_preview_length} bytes)"
            return_rows.append([Text(row_label, 'grey'), Text(stream_string, style)])

        add_preview_row(STREAM, stream_preview_string)
        add_preview_row(HEX, stream_preview_hex)
        return_rows.append([Text('StreamLength', style='grey'), size_text(len(self.stream_data))])
        return return_rows

    def _node_label(self) -> Text:
        text = Text('<', style='white')
        text.append(f'{self.idnum}', style='bright_white')
        text.append(':', style='white')
        text.append(self.colored_node_label())
        text.append('(', style='white')
        text.append(self.colored_node_type())
        text.append(')', style='white')
        text.append('>')
        return text

    def __rich__(self) -> Text:
        return self._node_label()[:-1] + self.colored_address(max_length=DEFAULT_MAX_ADDRESS_LENGTH) + Text('>')

    def __str__(self) -> str:
        return self._node_label().plain

    def __repr__(self) -> str:
        return self.__str__()


# TODO: This should probably live in pdf_object_helper.py but the use of PdfTreeNode() makes that a circular reference
def resolve_references(reference_key: str, obj: PdfObject):
    """Recursively build the same data structure except IndirectObjects are resolved to nodes"""
    if isinstance(obj, NumberObject):
        return obj.as_numeric()
    elif isinstance(obj, IndirectObject):
        return PdfTreeNode.from_reference(obj, reference_key)
    elif isinstance(obj, list):
        return [resolve_references(reference_key, item) for item in obj]
    elif isinstance(obj, dict):
        return {k: resolve_references(k, v) for k, v in obj.items()}
    else:
        return obj


def to_table_row(reference_key: str, obj: PdfObject, is_single_row_table=False) -> List[Union[Text, str]]:
    """Turns PDF object properties into a formatted 3-tuple for use in Rich tables representing PdfObjects"""
    # 3rd col (AKA type(value)) is redundant if it's something like a TextString or Number node so we set it to ''
    return [
        Text(f"{reference_key}", style='grey' if isinstance(reference_key, int) else ''),
        Text(f"{escape(str(resolve_references(reference_key, obj)))}", style=get_type_style(type(obj))),
        '' if is_single_row_table else Text(pypdf_class_name(obj), style=get_type_string_style(type(obj)))
    ]


def get_node_type_style(obj):
    klass_string = pypdf_class_name(obj)

    if 'Dictionary' in klass_string:
        style = TYPE_STYLES[dict]
    elif 'EncodedStream' in klass_string:
        style = YARALYZER_THEME.styles['bytes']
    elif 'Stream' in klass_string:
        style = YARALYZER_THEME.styles['bytes.title']
    elif 'Text' in klass_string:
        style = YARALYZER_THEME.styles['grey.light']
    elif 'Array' in klass_string:
        style = PDF_ARRAY
    else:
        style = 'bright_yellow'

    return f"{style} italic"



"""
Unify font information spread across a bunch of PdfObjects (Font, FontDescriptor,
and FontFile) into a single class.
"""
from typing import Union

from PyPDF2._cmap import build_char_map, prepare_cm
from PyPDF2.generic import IndirectObject, PdfObject
from rich.columns import Columns
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from yaralyzer.config import YaralyzerConfig
from yaralyzer.helpers.bytes_helper import print_bytes
from yaralyzer.output.rich_console import console
from yaralyzer.output.rich_layout_elements import bytes_hashes_table
from yaralyzer.util.logging import log

from pdfalyzer.binary.binary_scanner import BinaryScanner
from pdfalyzer.config import PdfalyzerConfig
from pdfalyzer.detection.yaralyzer_helper import get_bytes_yaralyzer
from pdfalyzer.output.layout import get_label_style
from pdfalyzer.helpers.rich_text_helper import get_type_style
from pdfalyzer.helpers.string_helper import pp
from pdfalyzer.output.layout import print_section_subheader, print_headline_panel, subheading_width
from pdfalyzer.util.adobe_strings import (FONT, FONT_DESCRIPTOR, FONT_FILE, FONT_LENGTHS, RESOURCES, SUBTYPE,
     TO_UNICODE, TYPE, W, WIDTHS)


CHARMAP_TITLE_PADDING = (1, 0, 0, 2)
CHARMAP_PADDING = (0, 2, 0, 10)
CHARMAP_TITLE = 'Character Mapping (As Extracted By PyPDF2)'
FONT_SECTION_PREVIEW_LEN = 30

ATTRIBUTES_TO_SHOW_IN_SUMMARY_TABLE = [
    'sub_type',
    'base_font',
    'flags',
    'bounding_box',
]


class FontInfo:
    @classmethod
    def extract_font_infos(cls, obj_with_resources: PdfObject) -> ['FontInfo']:
        """
        Extract all the fonts from a given /Resources PdfObject node.
        obj_with_resources must have '/Resources' because that's what _cmap module expects
        """
        resources = obj_with_resources[RESOURCES]

        if isinstance(resources, IndirectObject):
            resources = resources.get_object()

        fonts = resources.get(FONT)

        if fonts is None:
            log.info(f'No fonts found in {obj_with_resources}')
            return []

        fonts = fonts.get_object()
        return [cls.build(label, font, obj_with_resources) for label, font in fonts.items()]

    @classmethod
    def build(cls, label: str, font_ref: IndirectObject, obj_with_resources) -> 'FontInfo':
        """Build a FontInfo object from a IndirectObject ref to a /Font"""
        font_obj = font_ref.get_object()
        font_descriptor = None
        font_file = None

        if font_obj.get(TYPE) != FONT:
            raise TypeError(f"{TYPE} of {font_ref} is not {FONT}")

        if FONT_DESCRIPTOR in font_obj:
            font_descriptor = font_obj[FONT_DESCRIPTOR].get_object()
            font_file_keys = [k for k in font_descriptor.keys() if FONT_FILE in k]

            # There's /FontFile, /FontFile2, etc. but whichever it is there should be only one
            if len(font_file_keys) > 1:
                raise RuntimeError(f"Too many /FontFile keys in {font_descriptor}: {font_file_keys}")
            elif len(font_file_keys) == 0:
                log.info(f"No font_file found in {font_descriptor}")
            else:
                font_file = font_descriptor[font_file_keys[0]].get_object()

        return cls(label, font_ref.idnum, font_obj, font_descriptor, font_file, obj_with_resources)

    def __init__(self, label, idnum, font, font_descriptor, font_file, obj_with_resources):
        self.label = label
        self.idnum = idnum
        self.font_file = font_file
        self.descriptor = font_descriptor

        # /Font attributes
        self.font = font
        self.sub_type = font.get(SUBTYPE)
        self.widths = font.get(WIDTHS) or font.get(W)

        if isinstance(self.widths, IndirectObject):
            self.widths = self.widths.get_object()

        self.base_font = font.get('/BaseFont')
        self.first_and_last_char = [font.get('/FirstChar'), font.get('/LastChar')]
        self.display_title = f"{self.idnum}. Font {self.label} "

        if self.sub_type is None:
            log.warning(f"Font type not given for {self.display_title}")
            self.display_title += "(UNKNOWN FONT TYPE)"
        else:
            self.display_title += f"({self.sub_type[1:]})"

        # FontDescriptor attributes
        if font_descriptor is not None:
            self.bounding_box = font_descriptor.get('/FontBBox')
            self.flags = font_descriptor.get('/Flags')
        else:
            self.bounding_box = None
            self.flags = None

        # /FontFile attributes
        if font_file is not None:
            self.lengths = [font_file[k] for k in FONT_LENGTHS if k in font_file]
            self.stream_data = font_file.get_data()
            self.advertised_length = sum(self.lengths)
            scanner_label = Text(self.display_title, get_label_style(FONT_FILE))
            self.binary_scanner = BinaryScanner(self.stream_data, self, scanner_label)
            self.prepared_char_map = prepare_cm(font) if TO_UNICODE in font else None
            # TODO: shouldn't we be passing ALL the widths?
            self._char_map = build_char_map(label, self.widths[0], obj_with_resources)

            try:
                self.character_mapping = self._char_map[3]
            except (IndexError, TypeError):
                log.warning(f"Exception trying to get character mapping for {self}")
                self.character_mapping = []
        else:
            self.lengths = None
            self.stream_data = None
            self.advertised_length = None
            self.binary_scanner = None
            self.prepared_char_map = None
            self._char_map = None
            self.character_mapping = None

    def yara_scan(self) -> None:
        if self.binary_scanner is None:
            log.debug(f"No binary to scan for {self.display_title}")
            return

        get_bytes_yaralyzer(self.stream_data, str(self)).yaralyze()

    def width_stats(self):
        if self.widths is None:
            return {}

        return {
            'min': min(self.widths),
            'max': max(self.widths),
            'count': len(self.widths),
            'unique_count': len(set(self.widths)),
        }

    def print_summary(self):
        """Prints a table of info about the font drawn from the various PDF objects. quote_type of None means all."""
        print_section_subheader(str(self), style='font.title')
        #console.print(Panel(self.display_title, width=subheading_width(), padding=(1, 1)), style='font.title')
        console.print(self._summary_table())
        console.line()
        self.print_character_mapping()
        self.print_prepared_charmap()
        console.line()

    def print_character_mapping(self):
        """Prints the character mapping extracted by PyPDF2._charmap in tidy columns"""
        if self.character_mapping is None or len(self.character_mapping) == 0:
            log.info(f"No character map found in {self}")
            return

        print_headline_panel(f"{self} {CHARMAP_TITLE}", style='charmap.title')
        charmap_entries = [_format_charmap_entry(k, v) for k, v in self.character_mapping.items()]

        charmap_columns = Columns(
            charmap_entries,
            column_first=True,
            padding=CHARMAP_PADDING,
            equal=True,
            align='right')

        console.print(Padding(charmap_columns, CHARMAP_TITLE_PADDING), width=subheading_width())
        console.line()

    def print_prepared_charmap(self):
        """Prints the prepared_charmap returned by PyPDF2"""
        if self.prepared_char_map is None:
            log.info(f"No prepared_charmap found in {self}")
            return

        headline = f"{self} Adobe PostScript charmap prepared by PyPDF2"
        print_headline_panel(headline, style='charmap.prepared_title')
        print_bytes(self.prepared_char_map, style='charmap.prepared')
        console.print('')

    def preview_bytes_at_advertised_lengths(self):
        """Show the bytes at the boundaries provided by /Length1, /Length2, and /Length3, if they exist"""
        lengths = self.lengths or []

        if self.lengths is None or len(lengths) <= 1:
            console.print("No length demarcations to preview.", style='grey.dark')

        for i, demarcation in enumerate(lengths[1:]):
            console.print(f"{self.font_file} at /Length{i} ({demarcation}):")
            print(f"\n  Stream before: {self.stream_data[demarcation - FONT_SECTION_PREVIEW_LEN:demarcation + 1]}")
            print(f"\n  Stream after: {self.stream_data[demarcation:demarcation + FONT_SECTION_PREVIEW_LEN]}")

        print(f"\nfinal bytes back from {self.stream_data.lengths[2]} + 10: {self.stream_data[-10 - -f.lengths[2]:]}")

    def _summary_table(self):
        """Build a Rich Table with important info about the font"""
        table = Table('', '', show_header=False)
        table.columns[0].style = 'font.property'
        table.columns[0].justify = 'right'

        def add_table_row(name, value):
            table.add_row(name, Text(str(value), get_type_style(type(value))))

        for attr in ATTRIBUTES_TO_SHOW_IN_SUMMARY_TABLE:
            attr_value = getattr(self, attr)
            add_table_row(attr, attr_value)

        add_table_row('/Length properties', self.lengths)
        add_table_row('total advertised length', self.advertised_length)

        if self.binary_scanner is not None:
            add_table_row('actual length', self.binary_scanner.stream_length)

        if self.prepared_char_map is not None:
            add_table_row('prepared charmap length', len(self.prepared_char_map))

        if self._char_map is not None:
            add_table_row('character mapping count', len(self.character_mapping))

        if self.widths is not None:
            for k, v in self.width_stats().items():
                add_table_row(f"char width {k}", v)

        if self.widths is not None:
            if len(set(self.widths)) == 1:
                table.add_row(
                    'char widths',
                    Text(f"{self.widths[0]} (single value repeated {len(self.widths)} times)", style=get_type_style(list)))
            else:
                add_table_row('char widths', self.widths)
                add_table_row('char widths(sorted)', sorted(self.widths))

        col_0_width = max([len(entry) for entry in table.columns[0]._cells]) + 4
        table.columns[1].max_width = subheading_width() - col_0_width - 3
        return table

    def __str__(self) -> str:
        return self.display_title


def _format_charmap_entry(k, v):
    key = pp.pformat(k)

    for quote_char in ['"', "'"]:
        if len(key) > 1 and key.startswith(quote_char) and key.endswith(quote_char):
            key = key[1:-1]

    return quoted_text(key, 'charmap.byte') + Text(' => ') + quoted_text(str(v), 'charmap.char')


def quoted_text(_string, style: Union[str, None]=None, quote_char_style='white', quote_char="'") -> Text:
    quote_char = Text(quote_char, style=quote_char_style)
    txt = quote_char + Text(_string, style=style or '') + quote_char
    txt.justify = 'center'
    return txt
